from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.pieces.structs.aoe2_struct import AoE2Struct


class PlayerDataFourStruct(AoE2Struct):
    def __init__(self, parser_obj=None, data=None, pieces=None):
        retrievers = [
            Retriever("food_duplicate", DataType("f32")),
            Retriever("wood_duplicate", DataType("f32")),
            Retriever("gold_duplicate", DataType("f32")),
            Retriever("stone_duplicate", DataType("f32")),
            Retriever("ore_x_duplicate", DataType("f32")),
            Retriever("trade_goods_duplicate", DataType("f32")),
            Retriever("population_limit", DataType("f32")),
        ]

        super().__init__("Player Data #4", retrievers, parser_obj, data, pieces=pieces)

    @staticmethod
    def defaults(pieces):
        defaults = {
            'food_duplicate': 0.0,
            'wood_duplicate': 0.0,
            'gold_duplicate': 0.0,
            'stone_duplicate': 0.0,
            'ore_x_duplicate': 0.0,
            'trade_goods_duplicate': 0.0,
            'population_limit': 200.0,
        }
        return defaults



from typing import Dict

from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.helper.retriever_dependency import DependencyAction, RetrieverDependency, DependencyTarget, \
    DependencyEval
from AoE2ScenarioParser.pieces.structs.aoe2_struct import AoE2Struct
from AoE2ScenarioParser.pieces.structs.condition import ConditionStruct
from AoE2ScenarioParser.pieces.structs.effect import EffectStruct


class TriggerStruct(AoE2Struct):
    dependencies: Dict[str, Dict[str, RetrieverDependency]] = {
        "number_of_effects": {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_VALUE, DependencyTarget("self", "effect_data"),
                DependencyEval("len(x)")
            )},
        "effect_data": {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_REPEAT, DependencyTarget("self", "number_of_effects")
            ),
            "on_construct": RetrieverDependency(DependencyAction.REFRESH_SELF),
            "on_commit": RetrieverDependency(
                DependencyAction.REFRESH, DependencyTarget("self", "number_of_effects")
            )
        },
        "effect_display_order_array": {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_REPEAT, DependencyTarget("self", "number_of_effects")
            ),
            "on_construct": RetrieverDependency(DependencyAction.REFRESH_SELF)
        },
        "number_of_conditions": {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_VALUE, DependencyTarget("self", "condition_data"),
                DependencyEval("len(x)")
            )
        },
        "condition_data": {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_REPEAT, DependencyTarget("self", "number_of_conditions")
            ),
            "on_construct": RetrieverDependency(DependencyAction.REFRESH_SELF),
            "on_commit": RetrieverDependency(
                DependencyAction.REFRESH, DependencyTarget("self", "number_of_conditions")
            )
        },
        "condition_display_order_array": {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_REPEAT,
                DependencyTarget("self", "number_of_conditions")
            ),
            "on_construct": RetrieverDependency(DependencyAction.REFRESH_SELF)
        },
    }

    def __init__(self, parser_obj=None, data=None, pieces=None):
        retrievers = [
            Retriever("enabled", DataType("u32")),
            Retriever("looping", DataType("s8")),
            Retriever("description_string_table_id", DataType("s32")),
            Retriever("display_as_objective", DataType("u8")),
            Retriever("objective_description_order", DataType("u32")),
            Retriever("make_header", DataType("u8")),
            Retriever("short_description_string_table_id", DataType("s32")),
            Retriever("display_on_screen", DataType("u8")),
            Retriever("unknown", DataType("5")),
            Retriever("mute_objectives", DataType("u8")),
            # Retriever("Trigger starting time", DataType("u32")),
            Retriever("trigger_description", DataType("str32")),
            Retriever("trigger_name", DataType("str32")),  # (max 44 characters in UI)
            Retriever("short_description", DataType("str32")),
            Retriever("number_of_effects", DataType("s32")),
            Retriever("effect_data", DataType(EffectStruct)),
            Retriever("effect_display_order_array", DataType("s32")),
            Retriever("number_of_conditions", DataType("s32")),
            Retriever("condition_data", DataType(ConditionStruct)),
            Retriever("condition_display_order_array", DataType("s32"))
        ]

        super().__init__("Trigger", retrievers, parser_obj, data, pieces=pieces)

    @staticmethod
    def defaults(pieces):
        defaults = {
            'enabled': 1,
            'looping': 0,
            'description_string_table_id': -1,
            'display_as_objective': 0,
            'objective_description_order': 0,
            'make_header': 0,
            'short_description_string_table_id': -1,
            'display_on_screen': 0,
            'unknown': b'\x00\x00\x00\x00\x00',
            'mute_objectives': 0,
            'trigger_description': '',
            'trigger_name': 'Trigger 0',
            'short_description': '',
            'number_of_effects': 0,
            'effect_data': [],
            'effect_display_order_array': [],
            'number_of_conditions': 0,
            'condition_data': [],
            'condition_display_order_array': [],
        }
        return defaults



from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.pieces import aoe2_piece


class CinematicsPiece(aoe2_piece.AoE2Piece):
    def __init__(self, parser_obj=None, data=None, pieces=None):
        retrievers = [
            Retriever("ascii_pregame", DataType("str16")),
            Retriever("ascii_victory", DataType("str16")),
            Retriever("ascii_loss", DataType("str16")),
            # Retriever("Separator (! in some version)", DataType("1")),
        ]

        super().__init__("Cinematics", retrievers, parser_obj, data=data, pieces=pieces)

    @staticmethod
    def defaults(pieces):
        defaults = {
            'ascii_pregame': '',
            'ascii_victory': '',
            'ascii_loss': '',
        }
        return defaults



from AoE2ScenarioParser.objects.aoe2_object import AoE2Object


class DiplomacyObject(AoE2Object):
    def __init__(self,
                 player_stances
                 ):
        self.player_stances = player_stances

        super().__init__()



uncapped = [
    'if', 'a', 'is', 'by', 'of', 'the', 'to', 'per', 'in', 'an', 'for', 'and', 'from','does', 'not', 'also', 'have',
    'this', 'set', 'will'
]
full_capped = [
    'swgb', 'ptwc'
]

with open('../../../resources/personal_docs/ResourceAttributeDescriptionsV1.4.1.md', 'r') as file:
    state = 0
    state_lines = []
    file = file.read()
    for index, line in enumerate(file.splitlines()):
        if line.strip() == '' or line[0] == '`':
            if state == 1:
                # Add colon if first word is 'unknown' or 'boolean'
                words = state_lines[0].split(' ')
                if words[0].lower() in ['boolean', 'unknown', 'unused']:
                    words[0] = words[0] + ':'
                    state_lines[0] = ' '.join(words)
                if len(state_lines) > 1:  # Combine gathered list
                    if state_lines[0][-1] == ".":
                        state_lines[0] = state_lines[0][:-1]
                    state_lines[0] += ":"
                    state_lines.insert(1, "")
                    print('\n'.join(["\"\"\""] + state_lines + ["\"\"\""]))
                else:  # Print single desc line
                    print("\"\"\"" + state_lines[0] + "\"\"\"")
                state_lines = []
                state = 0
            if line.strip() == '':
                break

            # `-------- Building Limit = 30;`
            # BUILDING_LIMIT = 30
            if line[-2:] == ';`':
                line = line[:-2]
            elif line[-1] == '`':
                line = line[:-1]
                print(line)
                print(index)
                exit()
            splitted = line[1:].split(' = ')
            splitted[0] = splitted[0].replace('--------', '').replace('========', '').replace('~~~~~~~~', '')
            splitted[0] = splitted[0].replace('/', ' or ')
            splitted[0] = splitted[0].replace('(', ' ').replace(')', ' ').replace('  ', ' ').strip()
            if splitted[0][:6].lower() == 'unused':
                splitted[0] = f"UNUSED_RESOURCE_{splitted[1].zfill(3)}"
            print(f"{splitted[0].replace(' ', '_').upper()} = {splitted[1]}")
        elif line.strip()[0] == '<':
            state = 1

            # <BOOLEAN_ALLOW_ENEMY_BUILDING_CONVERSIONS by the source player.>
            # Boolean Allow Enemy Building Conversions by the Source Player:
            replaced_line = line.replace('_', ' ')
            edited_line = replaced_line.strip()[1:-1].split(' ')
            items = []
            for item in edited_line:
                lower_item = item.lower()
                if lower_item in uncapped:
                    items.append(item.lower())
                elif lower_item in full_capped:
                    items.append(item.upper())
                else:
                    items.append(item.capitalize())

            state_lines.append(' '.join(items))
        elif line.strip()[:2] == '**':
            state = 1

            # **0: when all techs not researched**
            #     0: When All Techs Not Researched
            replaced_line = line.strip()[:-2].replace('**', '- ')
            state_lines.append(' '.join(
                [item.lower() if item.lower() in uncapped else item.capitalize() for item in replaced_line.split(' ')]
            ))



from typing import Dict

from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.helper.retriever_dependency import RetrieverDependency, DependencyAction, DependencyTarget, \
    DependencyEval
from AoE2ScenarioParser.pieces import aoe2_piece
from AoE2ScenarioParser.pieces.structs.player_data_one import PlayerDataOneStruct


class DataHeaderPiece(aoe2_piece.AoE2Piece):
    dependencies: Dict[str, Dict[str, RetrieverDependency]] = {
        "mission_items_counter": {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_VALUE, DependencyTarget("self", "mission_item"),
                DependencyEval("len(x)"))
        },
        "mission_item": {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_REPEAT, DependencyTarget("self", "mission_items_counter")),
            "on_construct": RetrieverDependency(DependencyAction.REFRESH_SELF),
            "on_commit": RetrieverDependency(
                DependencyAction.REFRESH, DependencyTarget("self", "mission_items_counter"))
        },
        "unknown2": {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_REPEAT, DependencyTarget("FileHeaderPiece", "version"),
                DependencyEval("1 if x == '1.40' else 0")),
            "on_construct": RetrieverDependency(DependencyAction.REFRESH_SELF)
        },
    }

    def __init__(self, parser_obj=None, data=None, pieces=None):
        retrievers = [
            Retriever("next_unit_id_to_place", DataType("u32")),
            Retriever("version", DataType("f32")),
            Retriever("player_names", DataType("c256", repeat=16)),
            Retriever("string_table_player_names", DataType("u32", repeat=16)),
            Retriever("player_data_1", DataType(PlayerDataOneStruct, repeat=16)),
            Retriever("conquest_mode", DataType("u8")),
            Retriever("mission_items_counter", DataType("u16")),
            Retriever("mission_available", DataType("u16")),
            Retriever("mission_timeline", DataType("f32")),
            Retriever("mission_item", DataType("30")),
            Retriever("unknown", DataType("64")),
            Retriever("filename", DataType("str16")),
        ]

        super().__init__("Data Header", retrievers, parser_obj, data=data, pieces=pieces)

    @staticmethod
    def defaults(pieces):
        defaults = {
            'next_unit_id_to_place': 0,
            'version': 1.399999976158142,
            'player_names': ['\x00' * 256] * 16,
            'string_table_player_names': [4294967294] * 16,
            'player_data_1': DataHeaderPiece._get_player_data_1_default(pieces),
            'conquest_mode': 0,
            'mission_items_counter': 0,
            'mission_available': 0,
            'mission_timeline': 0.0,
            'mission_item': [],
            'unknown': b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
                       b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
                       b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00',
            'filename': 'CreatedUsingAoE2ScenarioParser.aoe2scenario',
        }
        return defaults

    @staticmethod
    def _get_player_data_1_default(pieces):
        # active, human, civilization, cty_mode
        data = [list(PlayerDataOneStruct.defaults(pieces).values()) for _ in range(2)]
        data[0][0] = 1  # P1 Active
        data[0][1] = 1  # P1 Human
        data[1][0] = 1  # P2 Active
        data += [list(PlayerDataOneStruct.defaults(pieces).values()) for _ in range(14)]
        return [PlayerDataOneStruct(data=x, pieces=pieces) for x in data]



# Parser

import time
from typing import Any, List

import AoE2ScenarioParser.pieces.structs.aoe2_struct
from AoE2ScenarioParser.helper.bytes_to_x import *
from AoE2ScenarioParser.helper.generator import repeat_generator as r_gen
from AoE2ScenarioParser.helper.retriever import Retriever, get_retriever_by_name
from AoE2ScenarioParser.helper.retriever_dependency import DependencyAction

types = [
    "s",  # Signed int
    "u",  # Unsigned int
    "f",  # FloatingPoint
    "c",  # Character string
    "str",  # Variable length string
    "data",  # Data (Can be changed by used using bytes_to_x functions)
]
attributes = ['on_refresh', 'on_construct', 'on_commit']


def vorl(var: Any, retriever: Retriever = None):
    """vorl stands for "Variable or List". This function returns the value if the list is a size of 1"""
    if Retriever is not None and retriever.possibly_list:
        dependencies = []
        for attribute in attributes:
            if hasattr(retriever, attribute):
                dependencies += listify(getattr(retriever, attribute))
                if retriever.datatype.repeat != 1 or DependencyAction.SET_REPEAT in [x.dependency_type for x in
                                                                                     dependencies]:
                    return listify(var)
    if type(var) is list:
        if len(var) == 1:
            return var[0]
    return var


def listify(var) -> list:
    """Always return item as list"""
    if type(var) is list:
        return var
    else:
        return [var]


class Parser:
    def __init__(self):
        self._saves = dict()

    def retrieve_value(self, generator, retriever, retrievers=None, pieces=None, as_length=False) -> Any:
        if (pieces is None or retrievers is None) and not as_length:
            raise ValueError("Normal retrieval of length requires pieces parameter.")
        length = 0
        result = list()
        var_type, var_len = datatype_to_type_length(retriever.datatype.var)

        if hasattr(retriever, 'on_construct'):
            handle_retriever_dependency(retriever, retrievers, "construct", pieces)

        try:
            for i in range(0, retriever.datatype.repeat):
                length += var_len

                if var_type == "struct":
                    val = retriever.datatype.var(self)
                    result.append(val)
                    val.set_data_from_generator(generator, pieces)
                    i = val.get_length()

                    length += i
                    continue
                if var_type == "u" or var_type == "s":
                    val = bytes_to_int(r_gen(generator, var_len), signed=(var_type == "s"))
                elif var_type == "f":
                    if var_len == 4:
                        val = bytes_to_float(r_gen(generator, var_len))
                    else:  # Always 4 except for trigger version
                        val = bytes_to_double(r_gen(generator, var_len))
                elif var_type == "c":
                    val = bytes_to_fixed_chars(r_gen(generator, var_len))
                elif var_type == "data":
                    val = r_gen(generator, var_len, intended_stop_iteration=(retriever.name == "__END_OF_FILE_MARK__"))
                elif var_type == "str":
                    string_length = bytes_to_int(r_gen(generator, var_len), endian="little", signed=True)
                    try:
                        data = r_gen(generator, string_length)
                        val = bytes_to_str(data)
                        length += string_length
                    except StopIteration as e:
                        print(f"\n[StopIteration] Parser.retrieve_value: \n"
                              f"\tRetriever: {retriever}\n"
                              f"\tString length: {string_length}\n")
                        raise e
                else:
                    break
                result.append(val)
        except StopIteration as e:
            if retriever.name == "__END_OF_FILE_MARK__":
                retriever.datatype.repeat = 0
                return True, None, None
            else:
                raise e
        except Exception as e:
            return vorl(result, retriever), None if not as_length else length, e

        # TODO: REMOVE THIS AFTER TRUE VERSION SUPPORT
        if retriever.name == "version" and retriever.datatype.var == "c4":
            version = vorl(result, retriever)
            if version not in ["1.40", "1.41"]:
                print("\n\n")
                print('\n'.join([
                    "#### SORRY FOR THE INCONVENIENCE ####",
                    f"Your current version is: '{version}'. ",
                    f"The only supported versions currently are: '1.40' & '1.41'.",
                    f"If this scenario has a version below '1.40', please continue reading below. If your version is "
                    f"higher, I'm probably working on supporting it. Hopefully within a couple of days!",
                    "The reason for this is a huge rework for version support.",
                    "This rework will take some time to complete, so until then, please upgrade your scenario to the "
                    "newest version. You can do this by saving it again in the in-game editor.",
                    "If you do not want to upgrade the scenarios, please downgrade this library to version 0.0.11. You "
                    "can do so by executing the following command in cmd:",
                    "",
                    ">>> pip install --force-reinstall AoE2ScenarioParser==0.0.11",
                    "",
                    "Thank you in advance."
                ]))
                time.sleep(1)
                print("- KSneijders")
                print("\n\n")
                time.sleep(1)
                raise ValueError("Currently unsupported version. Please read the message above. Thank you.")
        elif retriever.name == "__END_OF_FILE_MARK__":
            result = b''
            try:
                while True:
                    result += next(generator)
            except StopIteration:
                pass
            print("\n\n" + "\n".join([
                "The file being read has more bytes than anticipated.",
                "Please notify me (MrKirby/KSneijders) about this message!",
                "This will help with understanding more parts of scenario files! Thanks in advance!",
                "You can contact me using:",
                "- Discord: MrKirby # 5063",
                "- Github: https://github.com/KSneijders/AoE2ScenarioParser/issues",
                "",
                "Please be so kind and include the map in question. Thanks again!\n\n",
                "",
                "Extra data found in the file:",
                f"\t'{result}'"
            ]))
            retriever.datatype.repeat = 1
            return result, None, None

        return vorl(result, retriever), None if not as_length else length, None


def handle_retriever_dependency(retriever: Retriever, retrievers: List[Retriever], state, pieces):
    a = 1 if retriever.name == "image" else 0
    if state == "construct":
        retriever_on_x = retriever.on_construct
    elif state == "commit":
        retriever_on_x = retriever.on_commit
    elif state == "refresh":
        retriever_on_x = retriever.on_refresh
    else:
        raise ValueError("State must be any of: construct, commit or refresh")

    retriever_on_x_list = listify(retriever_on_x)
    for retriever_on_x in retriever_on_x_list:
        dep_action = retriever_on_x.dependency_type
        dep_target = retriever_on_x.dependency_target
        if dep_action == DependencyAction.REFRESH_SELF:
            handle_retriever_dependency(retriever, retrievers, "refresh", pieces)
        elif dep_action == DependencyAction.REFRESH:
            listified_target = listify(dep_target.target_piece)
            listified_target_attr = listify(dep_target.piece_attr_name)
            for i in range(len(listified_target)):
                retriever_list = handle_dependency_target(listified_target[i], retrievers, pieces)
                retriever_to_be_refreshed = get_retriever_by_name(retriever_list, listified_target_attr[i])
                handle_retriever_dependency(retriever_to_be_refreshed, retriever_list, "refresh", pieces)
        elif dep_action in [DependencyAction.SET_VALUE, DependencyAction.SET_REPEAT]:
            listified_target = listify(dep_target.target_piece)
            listified_target_attr = listify(dep_target.piece_attr_name)
            retriever_data = []
            for i in range(len(listified_target)):
                retriever_list = handle_dependency_target(listified_target[i], retrievers, pieces)
                retriever_data.append(get_retriever_by_name(retriever_list, listified_target_attr[i]).data)
            value = handle_dependency_eval(retriever_on_x, retriever_data)
            if dep_action == DependencyAction.SET_VALUE:
                retriever.data = value
            elif dep_action == DependencyAction.SET_REPEAT:
                retriever.datatype.repeat = value


def handle_dependency_target(target_piece, retrievers, pieces):
    if target_piece == "self":
        retriever_list = retrievers
    else:
        retriever_list = pieces[target_piece].retrievers
    return retriever_list


def handle_dependency_eval(retriever_on_x, value):
    a = 1 if retriever_on_x.dependency_eval.eval_code == "width * height" else 0
    eval_locals = retriever_on_x.dependency_eval.eval_locals
    values_as_variable = retriever_on_x.dependency_eval.values_as_variable
    # If value as is used, use it as keys for the value!
    if values_as_variable:
        eval_locals = {**eval_locals, **dict(zip(values_as_variable, value))}
    else:
        eval_locals['x'] = value[0]
    return eval(retriever_on_x.dependency_eval.eval_code, {}, eval_locals)


def parse_repeat_string(saves, repeat_string):
    while True:
        start = repeat_string.find("{")
        end = repeat_string.find("}")

        if start == -1 and end == -1:
            break

        inclusive = repeat_string[start:end + 1]
        exclusive = repeat_string[start + 1:end]

        repeat_string = repeat_string.replace(inclusive, str(saves[exclusive]))
    return eval(repeat_string)


def calculate_length(generator, retriever_list):
    parser = Parser()
    total_length = 0

    for retriever in retriever_list:
        result, length, status = parser.retrieve_value(generator, retriever, retriever_list, as_length=True)
        retriever.data = result
        total_length += length

    return total_length


def datatype_to_type_length(var):
    try:
        if issubclass(var, AoE2ScenarioParser.pieces.structs.aoe2_struct.AoE2Struct):
            return "struct", 0
        else:  # Not possible at this time
            return "", 0
    except TypeError:
        pass

    var_type = ""
    var_len = ""

    for char in var:
        if char.isnumeric():
            var_len += char
        else:
            var_type += char

    if var_type == "":
        var_type = "data"

    if var_len == "":
        var_len = 0
    else:
        var_len = int(var_len)

    assert var_type in types

    if var_type != "c" and var_type != "data":
        var_len = int(var_len / 8)

    return var_type, var_len


def retriever_to_bytes(retriever, pieces):
    var_type, var_len = datatype_to_type_length(retriever.datatype.var)

    return_bytes = b''

    is_list = type(retriever.data) == list
    if is_list:
        retriever.datatype.repeat = len(retriever.data)

    try:
        for i in range(0, retriever.datatype.repeat):
            data = retriever.data[i] if is_list else retriever.data

            if data is None:
                # No data is found in struct. Reasoning described below.
                return None

            if var_type == "struct":
                for struct_retriever in data.retrievers:
                    result = retriever_to_bytes(struct_retriever, pieces)
                    if result is None:
                        # Return default value. When non is committed.
                        # Should only happen when a value is not transferred from and to a struct.
                        # This is because structs are recreated on file generation. When the struct does not contain
                        # a certain value because it's use is unknown, the value isn't transferred between.
                        struct_retriever.data = retriever.datatype.var.defaults(pieces)[struct_retriever.name]
                        return_bytes += retriever_to_bytes(struct_retriever, pieces)
                        continue
                    return_bytes += result
            if var_type == "u" or var_type == "s":  # int
                return_bytes += int_to_bytes(data, var_len, signed=(var_type == "s"))
            elif var_type == "f":  # float
                if var_len == 4:
                    return_bytes += float_to_bytes(data)
                else:
                    return_bytes += double_to_bytes(data)
            elif var_type == "c":  # str
                return_bytes += fixed_chars_to_bytes(data)
            elif var_type == "data":  # bytes
                return_bytes += data
            elif var_type == "str":  # str
                byte_string = str_to_bytes(data, retriever)
                return_bytes += int_to_bytes(len(byte_string), var_len, endian="little", signed=True)
                return_bytes += byte_string
    except (AttributeError, TypeError) as e:
        data_text = repr(retriever.data)
        if type(retriever.data) == list and len(retriever.data) > 5:
            data_text = f"[{retriever.data[0].__class__.__name__}] * {len(retriever.data)}"

        print(f"\n{type(e).__name__} occurred in: {retriever.name} "
              f"\n\tData: {data_text}"
              f"\n\tDatatype: {str(retriever.datatype)}")
        raise e

    if retriever.log_value:
        print(retriever, "returned", return_bytes)

    return return_bytes



from typing import Dict

from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.helper.retriever_dependency import RetrieverDependency, DependencyTarget, DependencyAction, \
    DependencyEval
from AoE2ScenarioParser.pieces import aoe2_piece


def _fill_dependencies(dependencies):
    for disabled_type in ["tech", "building", "unit"]:
        dependencies[f"per_player_number_of_disabled_{disabled_type}s"] = {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_VALUE,
                DependencyTarget(["self"] * 8,
                                 [f"disabled_{disabled_type}_ids_player_{p}" for p in range(1, 9)]),
                DependencyEval("[len(x) for x in [p1, p2, p3, p4, p5, p6, p7, p8]]",
                               values_as_variable=['p1', 'p2', 'p3', 'p4', 'p5', 'p6', 'p7', 'p8'])
            )
        }
        for player in range(1, 9):
            dependencies[f"disabled_{disabled_type}_ids_player_{player}"] = {
                "on_refresh": RetrieverDependency(
                    DependencyAction.SET_REPEAT,
                    DependencyTarget("self", f"per_player_number_of_disabled_{disabled_type}s"),
                    DependencyEval(f"x[{player - 1}]")),
                "on_construct": RetrieverDependency(DependencyAction.REFRESH_SELF),
                "on_commit": RetrieverDependency(
                    DependencyAction.REFRESH,
                    DependencyTarget("self", f"per_player_number_of_disabled_{disabled_type}s")
                ) if player == 1 else None
            }


class OptionsPiece(aoe2_piece.AoE2Piece):
    dependencies: Dict[str, Dict[str, RetrieverDependency]] = {
        'number_of_triggers': {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_VALUE, DependencyTarget("TriggerPiece", "trigger_data"),
                DependencyEval("len(x)")
            )
        }
    }
    _fill_dependencies(dependencies)

    def __init__(self, parser_obj=None, data=None, pieces=None):
        retrievers = []
        for disabled_type in ["tech", "building", "unit"]:
            retrievers.append(Retriever(f"per_player_number_of_disabled_{disabled_type}s", DataType("u32", repeat=16)))
            for player in range(1, 9):
                retrievers.append(Retriever(f"disabled_{disabled_type}_ids_player_{player}", DataType("u32")))
                # Unused: Players 9 - 16 can't have {disabled_type} technologies
            retrievers.append(Retriever(f"disabled_{disabled_type}_ids_player9-16", DataType("u32", repeat=0)))

        retrievers += [
            Retriever("combat_mode", DataType("u32")),
            Retriever("naval_mode", DataType("u32")),
            Retriever("all_techs", DataType("u32")),
            Retriever("per_player_starting_age", DataType("u32", repeat=16)),
            # 2: Dark 6 = Post | 1-8 players 9 GAIA
            Retriever("unknown", DataType("32")),
            Retriever("number_of_triggers", DataType("u32")),
        ]
        super().__init__("Options", retrievers, parser_obj, data=data, pieces=pieces)

    @staticmethod
    def defaults(pieces):
        defaults = {}
        for disabled_type in ["tech", "building", "unit"]:
            defaults[f'per_player_number_of_disabled_{disabled_type}s'] = [0] * 16
            for player in range(1, 9):
                defaults[f'disabled_{disabled_type}_ids_player_{player}'] = []
            defaults[f'disabled_{disabled_type}_ids_player9-16'] = []

        defaults = dict(**defaults, **{
            'combat_mode': 0,
            'naval_mode': 0,
            'all_techs': 0,
            'per_player_starting_age': [2] * 16,
            # 'unknown': b'\x9d\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\t\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
            #            b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00',
            # 'unknown': b'\x9d\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\t\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
            #            b'\x00\x00\x00\x00\x00\x00\x00\x00\x00',
            'unknown': b'\x9d\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\r\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
                       b'\x00\x00\x00\x00\x00\x00\x00\x00\x00',

            'number_of_triggers': 0,
        })
        return defaults



from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.helper.retriever_dependency import RetrieverDependency, DependencyAction, DependencyTarget, \
    DependencyEval
from AoE2ScenarioParser.pieces import aoe2_piece
from AoE2ScenarioParser.pieces.structs.bitmap_info import BitMapInfoStruct


class BackgroundImagePiece(aoe2_piece.AoE2Piece):
    dependencies = {
        'bitmap_info': {
            "on_construct": RetrieverDependency(
                DependencyAction.SET_REPEAT, DependencyTarget("self", "picture_orientation"),
                DependencyEval("1 if x in [-1, 2] else 0")
            )
        },
    }

    def __init__(self, parser_obj=None, data=None, pieces=None):
        retrievers = [
            Retriever("ascii_filename", DataType("str16")),
            Retriever("picture_version", DataType("u32")),
            Retriever("bitmap_width", DataType("u32")),
            Retriever("bitmap_height", DataType("s32")),
            Retriever("picture_orientation", DataType("s16")),
            Retriever("bitmap_info", DataType(BitMapInfoStruct), possibly_list=False),
        ]

        super().__init__("Background Image", retrievers, parser_obj, data=data, pieces=pieces)

    @staticmethod
    def defaults(pieces):
        defaults = {
            'ascii_filename': '',
            'picture_version': 3,
            'bitmap_width': 0,
            'bitmap_height': 0,
            'picture_orientation': 1,
            'bitmap_info': []
        }
        return defaults



from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.pieces.structs.aoe2_struct import AoE2Struct


class ResourcesStruct(AoE2Struct):
    def __init__(self, parser_obj=None, data=None, pieces=None):
        retrievers = [
            Retriever("gold", DataType("u32")),
            Retriever("wood", DataType("u32")),
            Retriever("food", DataType("u32")),
            Retriever("stone", DataType("u32")),
            Retriever("ore_x_unused", DataType("u32")),
            Retriever("trade_goods", DataType("u32")),
            Retriever("player_color", DataType("u32"))
        ]

        super().__init__("Resources", retrievers, parser_obj, data, pieces=pieces)

    @staticmethod
    def defaults(pieces):
        defaults = {
            'gold': 0,
            'wood': 0,
            'food': 0,
            'stone': 0,
            'ore_x_unused': 0,
            'trade_goods': 0,
            'player_color': 0,
        }
        return defaults



from __future__ import annotations

import copy
from enum import IntEnum
from typing import List, Dict

from AoE2ScenarioParser.datasets.effects import Effect
from AoE2ScenarioParser.datasets.players import Player
from AoE2ScenarioParser.helper import helper
from AoE2ScenarioParser.helper.retriever_object_link import RetrieverObjectLink
from AoE2ScenarioParser.objects.aoe2_object import AoE2Object
from AoE2ScenarioParser.objects.trigger_obj import TriggerObject
from AoE2ScenarioParser.objects.variable_obj import VariableObject


class TriggersObject(AoE2Object):
    """Manager of the everything trigger related."""

    _link_list = [
        RetrieverObjectLink("triggers", "TriggerPiece", "trigger_data", process_as_object=TriggerObject),
        RetrieverObjectLink("trigger_display_order", "TriggerPiece", "trigger_display_order_array"),
        RetrieverObjectLink("variables", "TriggerPiece", "variable_data", process_as_object=VariableObject),
    ]

    def __init__(self,
                 triggers: List[TriggerObject],
                 trigger_display_order: List[int],
                 variables: List[VariableObject]
                 ):

        self._trigger_hash = helper.hash_list(triggers)
        self.triggers: List[TriggerObject] = triggers
        self.trigger_display_order: List[int] = trigger_display_order
        self.variables: List[VariableObject] = variables

        super().__init__()

    @property
    def trigger_display_order(self):
        if helper.list_changed(self.triggers, self._trigger_hash):
            helper.update_order_array(self._trigger_display_order, len(self.triggers))
            self._trigger_hash = helper.hash_list(self.triggers)
        return self._trigger_display_order

    @trigger_display_order.setter
    def trigger_display_order(self, val):
        self._trigger_display_order = val

    def copy_trigger_per_player(self,
                                from_player,
                                trigger_select,
                                change_from_player_only=False,
                                include_player_source=True,
                                include_player_target=False,
                                trigger_ce_lock=None,
                                include_gaia: bool = False,
                                create_copy_for_players: List[IntEnum] = None) -> Dict[Player, TriggerObject]:
        """
        Copies a trigger for all or a selection of players. Every copy will change desired player attributes with it.

        Args:
            from_player (IntEnum): The central player this trigger is created for. This is the player that will not get
                a copy.
            trigger_select (TriggerSelect): An object used to identify which trigger to select.
            change_from_player_only (bool): If set to True, only change player attributes in effects and conditions that
                are equal to the player defined using the `from_player` parameter.
            include_player_source (bool): If set to True, allow player source attributes to be changed while copying.
                Player source attributes are attributes where a player is defined to perform an action such as create an
                object. If set to False these attributes will remain unchanged.
            include_player_target (bool): If set to True, allow player target attributes to be changed while copying.
                Player target attributes are attributes where a player is defined as the target such as change ownership
                or sending resources. If set to False these attributes will remain unchanged.
            trigger_ce_lock (TriggerCELock): The TriggerCELock object. Used to lock certain (types) of conditions or
                effects from being changed while copying.
            include_gaia (bool): If True, GAIA is included in the copied list. (Also when `create_copy_for_players` is
                defined)
            create_copy_for_players (List[IntEnum]): A list of Players to create a copy for. The `from_player` will be
                excluded from this list.

        Returns:
            A dict with all the new created triggers. The key is the player for which the trigger is
                created using the IntEnum associated with it. Example:
                {Player.TWO: TriggerObject, Player.FIVE: TriggerObject}

        Raises:
            ValueError: if more than one trigger selection is used. Any of (trigger_index, display_index or trigger)
                Or if Both `include_player_source` and `include_player_target` are `False`

        :Authors:
            KSneijders

        """
        trigger_index, display_index, trigger = self._validate_and_retrieve_trigger_info(trigger_select)
        if not include_player_source and not include_player_target:
            raise ValueError("Cannot exclude player source and target.")

        if create_copy_for_players is None:
            create_copy_for_players = [
                Player.ONE, Player.TWO, Player.THREE, Player.FOUR,
                Player.FIVE, Player.SIX, Player.SEVEN, Player.EIGHT
            ]
        if include_gaia and Player.GAIA not in create_copy_for_players:
            create_copy_for_players.append(Player.GAIA)

        alter_conditions, alter_effects = TriggersObject._find_alterable_ce(trigger, trigger_ce_lock)

        return_dict: Dict[Player, TriggerObject] = {}
        for player in create_copy_for_players:
            if not player == from_player:
                new_trigger = self.copy_trigger(TS.trigger(trigger))
                new_trigger.name += f" (p{player})"
                return_dict[player] = new_trigger

                for cond_x in alter_conditions:
                    cond = new_trigger.conditions[cond_x]
                    # Player not set
                    if cond.source_player == -1:
                        continue
                    # Player not equal to 'from_player'
                    if change_from_player_only:
                        if not cond.source_player == from_player:
                            continue
                    # Change source player
                    if include_player_source:
                        cond.source_player = Player(player)
                    # Change target player
                    if include_player_target:
                        cond.target_player = Player(player)
                for effect_x in alter_effects:
                    effect = new_trigger.effects[effect_x]
                    # Player not set
                    if effect.source_player == -1:
                        continue
                    # Player not equal to 'from_player'
                    if change_from_player_only:
                        if not effect.source_player == from_player:
                            continue
                    # Change source player
                    if include_player_source:
                        effect.source_player = Player(player)
                    # Change target player
                    if include_player_target:
                        effect.target_player = Player(player)

        return return_dict

    def copy_trigger(self, trigger_select) -> TriggerObject:
        """
        Creates an exact copy (deepcopy) of this trigger.

        Args:
            trigger_select (TriggerSelect): An object used to identify which trigger to select.

        Returns:
            The newly copied trigger
        """
        trigger_index, display_index, trigger = self._validate_and_retrieve_trigger_info(trigger_select)

        deepcopy_trigger = copy.deepcopy(trigger)
        deepcopy_trigger.name += " (copy)"
        deepcopy_trigger.trigger_id = len(self.triggers)
        self.triggers.append(deepcopy_trigger)
        # helper.update_order_array(self.trigger_display_order, len(self.triggers))

        return deepcopy_trigger

    def copy_trigger_tree_per_player(self,
                                     from_player,
                                     trigger_select,
                                     change_from_player_only=False,
                                     include_player_source=True,
                                     include_player_target=False,
                                     trigger_ce_lock=None,
                                     include_gaia: bool = False,
                                     create_copy_for_players: List[IntEnum] = None,
                                     group_triggers_by=None):
        """
        Copies an entire trigger tree for all or a selection of players. Every copy will change desired player
        attributes with it. Trigger trees are triggers linked together using Effect.(DE)ACTIVATE_TRIGGER.

        Args:
            from_player (IntEnum): The central player this trigger is created for. This is the player that will not get
                a copy.
            trigger_select (TriggerSelect): An object used to identify which trigger to select.
            change_from_player_only (bool): If set to True, only change player attributes in effects and conditions that
                are equal to the player defined using the `from_player` parameter.
            include_player_source (bool): If set to True, allow player source attributes to be changed while copying.
                Player source attributes are attributes where a player is defined to perform an action such as create an
                object. If set to False these attributes will remain unchanged.
            include_player_target (bool): If set to True, allow player target attributes to be changed while copying.
                Player target attributes are attributes where a player is defined as the target such as change ownership
                or sending resources. If set to False these attributes will remain unchanged.
            trigger_ce_lock (TriggerCELock): The TriggerCELock object. Used to lock certain (types) of conditions or
                effects from being changed while copying.
            include_gaia (bool): If True, GAIA is included in the copied list. (Also when `create_copy_for_players` is
                defined)
            create_copy_for_players (List[IntEnum]): A list of Players to create a copy for. The `from_player` will be
                excluded from this list.
            group_triggers_by (GroupBy): How to group the newly added triggers.

        Returns:
            The newly created triggers in a dict using the Player as key and as value with a list of triggers
        """
        if group_triggers_by is None:
            group_triggers_by = GroupBy.NONE

        trigger_index, display_index, source_trigger = self._validate_and_retrieve_trigger_info(trigger_select)

        known_node_indexes = [trigger_index]
        self._find_trigger_tree_nodes_recursively(source_trigger, known_node_indexes)

        new_triggers = {}
        id_swap = {}
        for index in known_node_indexes:
            triggers = self.copy_trigger_per_player(
                from_player,
                TS.index(index),
                change_from_player_only,
                include_player_source,
                include_player_target,
                trigger_ce_lock,
                include_gaia,
                create_copy_for_players,
            )
            for player, trigger in triggers.items():
                id_swap.setdefault(index, {})[player] = trigger.trigger_id
                new_triggers.setdefault(player, []).append(trigger)

        for player, triggers in new_triggers.items():
            for trigger in triggers:
                activation_effects = [
                    effect for effect in trigger.effects if
                    effect.effect_type in [Effect.ACTIVATE_TRIGGER, Effect.DEACTIVATE_TRIGGER]
                ]
                for effect in activation_effects:
                    effect.trigger_id = id_swap[effect.trigger_id][player]

        # Group by logic
        if group_triggers_by == GroupBy.TRIGGER:
            for index, source_trigger_id in enumerate(known_node_indexes):
                for player, trigger in [(player, triggers[index]) for player, triggers in new_triggers.items()]:
                    # When going negative (going 'below' the source already happens at insert @ 0
                    display_index_offset = player - from_player if from_player <= player else 0
                    source_trigger_display_index = self.trigger_display_order.index(source_trigger_id)
                    self.trigger_display_order.remove(trigger.trigger_id)
                    new_display_index = max(0, source_trigger_display_index + display_index_offset)
                    self.trigger_display_order.insert(
                        new_display_index,
                        trigger.trigger_id
                    )
        elif group_triggers_by == GroupBy.PLAYER:
            source_trigger_display_index = display_index
            source_trigger_offset = 0
            # Group known tree nodes
            for tree_trigger_id in known_node_indexes:
                self.trigger_display_order.remove(tree_trigger_id)
                self.trigger_display_order.insert(
                    source_trigger_display_index + source_trigger_offset,
                    tree_trigger_id
                )
                source_trigger_offset += 1
            # Group copied triggers
            for player, triggers in new_triggers.items():
                source_trigger_display_index = self.trigger_display_order.index(source_trigger.trigger_id)
                source_trigger_offset = 0
                display_index_offset = (player - from_player if from_player <= player else 0) * len(known_node_indexes)
                for trigger in triggers:
                    final_offset = max(source_trigger_display_index + display_index_offset + source_trigger_offset, 0)
                    self.trigger_display_order.remove(trigger.trigger_id)
                    self.trigger_display_order.insert(
                        final_offset,
                        trigger.trigger_id
                    )
                    source_trigger_offset += 1

        return new_triggers

    def copy_trigger_tree(self, trigger_select: TriggerSelect) -> List[TriggerObject]:
        trigger_index, display_index, trigger = self._validate_and_retrieve_trigger_info(trigger_select)

        known_node_indexes = [trigger_index]
        self._find_trigger_tree_nodes_recursively(trigger, known_node_indexes)

        new_triggers = []
        id_swap = {}
        for index in known_node_indexes:
            trigger = self.copy_trigger(TS.index(index))
            new_triggers.append(trigger)
            id_swap[index] = trigger.trigger_id

        for trigger in new_triggers:
            activation_effects = [
                effect for effect in trigger.effects if
                effect.effect_type in [Effect.ACTIVATE_TRIGGER, Effect.DEACTIVATE_TRIGGER]
            ]
            for effect in activation_effects:
                effect.trigger_id = id_swap[effect.trigger_id]

        return new_triggers

    def replace_player(self, trigger_select, to_player, only_change_from=None, include_player_source=True,
                       include_player_target=False, trigger_ce_lock=None) -> TriggerObject:
        """
        Replaces player attributes. Specifically useful if multiple players are used in the same trigger.

        Args:
            trigger_select (TriggerSelect): An object used to identify which trigger to select.
            to_player (Player): The player the attributes are changed to.
            only_change_from (Player): Can only change player attributes if the player is equal to the given value
            include_player_source (bool): If set to True, allow player source attributes to be changed while replacing.
                Player source attributes are attributes where a player is defined to perform an action such as create an
                object. If set to False these attributes will remain unchanged.
            include_player_target (bool): If set to True, allow player target attributes to be changed while replacing.
                Player target attributes are attributes where a player is defined as the target such as change ownership
                or sending resources. If set to False these attributes will remain unchanged.
            trigger_ce_lock (TriggerCELock): The TriggerCELock object. Used to lock certain (types) of conditions or
                effects from being changed.

        Returns:
            The given trigger with the proper player attributes changed
        """
        trigger_index, display_index, trigger = self._validate_and_retrieve_trigger_info(trigger_select)
        alter_conditions, alter_effects = TriggersObject._find_alterable_ce(trigger, trigger_ce_lock)

        for cond_x in alter_conditions:
            cond = trigger.conditions[cond_x]
            if not cond.source_player == -1 and include_player_source:
                if only_change_from is not None and only_change_from != cond.source_player:
                    continue
                cond.source_player = Player(to_player)
            if not cond.target_player == -1 and include_player_target:
                if only_change_from is not None and only_change_from != cond.target_player:
                    continue
                cond.target_player = Player(to_player)
        for effect_x in alter_effects:
            effect = trigger.effects[effect_x]
            if not effect.source_player == -1 and include_player_source:
                if only_change_from is not None and only_change_from != effect.source_player:
                    continue
                effect.source_player = Player(to_player)
            if not effect.target_player == -1 and include_player_target:
                if only_change_from is not None and only_change_from != effect.target_player:
                    continue
                effect.target_player = Player(to_player)

        return trigger

    def add_trigger(self, name, description=None, description_stid=None, display_as_objective=None,
                    short_description=None, short_description_stid=None, display_on_screen=None, description_order=None,
                    enabled=None, looping=None, header=None, mute_objectives=None, conditions=None,
                    effects=None) -> TriggerObject:
        """
        Adds a new trigger to the scenario.

        Args:
            name (str): The name for the trigger
            description (str): The trigger description
            description_stid (int): The trigger description string table ID
            display_as_objective (bool): Display the trigger as objective
            short_description (str): The short trigger description
            short_description_stid (int): The short trigger description string table ID
            display_on_screen (bool): Display the trigger objective on screen
            description_order (int): ?
            enabled (bool): If the trigger is enabled from the start.
            looping (bool): If the trigger loops.
            header (bool): Turn objective into header
            mute_objectives (bool): Mute objectives
            conditions (List): A list of condition objects
            effects (List): A list of effect objects

        Returns:
            The newly created trigger

        """
        keys = [
            'description', 'description_stid', 'display_as_objective', 'short_description',
            'short_description_stid', 'display_on_screen', 'description_order', 'enabled', 'looping', 'header',
            'mute_objectives', 'conditions', 'effects'
        ]
        trigger_attr = {}
        for key in keys:
            if locals()[key] is not None:
                trigger_attr[key] = locals()[key]
        new_trigger = TriggerObject(name=name, trigger_id=len(self.triggers), **trigger_attr)
        self.triggers.append(new_trigger)
        # helper.update_order_array(self._trigger_display_order, len(self.triggers))
        return new_trigger

    def add_variable(self, name: str, variable_id: int = -1) -> VariableObject:
        """
        Adds a variable.

        Args:
            name (str): The name for the variable
            variable_id (int): The ID of the variable. If left empty (default: -1), lowest available value will be used

        Returns:
            The newly added VariableObject
        """
        list_of_var_ids = [var.variable_id for var in self.variables]
        if variable_id == -1:
            for i in range(256):
                if i not in list_of_var_ids:
                    variable_id = i
                    break
            if variable_id == -1:
                raise IndexError(f"No variable ID available. All in use? In use: ({list_of_var_ids}/256)")
        if not (0 <= variable_id <= 255):
            raise ValueError("Variable ID has to fall between 0 and 255 (incl).")
        if variable_id in list_of_var_ids:
            raise ValueError("Variable ID already in use.")

        new_variable = VariableObject(variable_id=variable_id, name=name)
        self.variables.append(new_variable)
        return new_variable

    def get_trigger(self, trigger_select: TriggerSelect) -> TriggerObject:
        trigger_index, display_index, trigger = self._validate_and_retrieve_trigger_info(trigger_select)
        return trigger

    def get_variable(self, variable_id: int = None, variable_name: str = None) -> VariableObject:
        if variable_id is None and variable_name is None:
            raise ValueError("Select a variable using the variable_id or variable_name parameters")
        if variable_id is not None and variable_name is not None:
            raise ValueError("Select a variable using either the variable_id or variable_name parameters, not both.")

        for variable in self.variables:
            if variable.variable_id == variable_id or variable.name == variable_name:
                return variable

    def remove_trigger(self, trigger_select: TriggerSelect) -> None:
        trigger_index, display_index, trigger = self._validate_and_retrieve_trigger_info(trigger_select)

        for x in self.trigger_display_order:
            if x > trigger_index:
                self.get_trigger(TS.index(x)).trigger_id -= 1

        del self.triggers[trigger_index]
        del self.trigger_display_order[display_index]

        self.trigger_display_order = [x - 1 if x > trigger_index else x for x in self.trigger_display_order]

    def _find_trigger_tree_nodes_recursively(self, trigger, known_node_indexes: List[int]) -> None:
        found_node_indexes = TriggersObject._find_trigger_tree_nodes(trigger)
        # unknown_node_indexes = found_node_indexes.difference(known_node_indexes)
        unknown_node_indexes = [i for i in found_node_indexes if i not in known_node_indexes]

        if len(unknown_node_indexes) == 0:
            return

        known_node_indexes += unknown_node_indexes

        for index in unknown_node_indexes:
            self._find_trigger_tree_nodes_recursively(self.triggers[index], known_node_indexes)

    def _validate_and_retrieve_trigger_info(self, trigger_select) -> (int, int, TriggersObject):
        trigger = trigger_select.trigger
        trigger_index = trigger_select.trigger_index
        display_index = trigger_select.display_index

        if trigger is not None:
            trigger_index = trigger.trigger_id
            display_index = self.trigger_display_order.index(trigger_index)
        elif trigger_index is not None:
            trigger = self.triggers[trigger_index]
            display_index = self.trigger_display_order.index(trigger_index)
        elif display_index is not None:
            trigger_index = self.trigger_display_order[display_index]
            trigger = self.triggers[trigger_index]

        return trigger_index, display_index, trigger

    def get_summary_as_string(self) -> str:
        return_string = "\nTrigger Summary:\n"

        triggers = self.triggers
        display_order = self.trigger_display_order

        if len(display_order) == 0:
            return_string += "\t<< No Triggers >>"

        longest_trigger_name = -1
        longest_index_notation = -1
        for display, trigger_index in enumerate(display_order):
            trigger_name = triggers[trigger_index].name
            longest_trigger_name = max(longest_trigger_name, len(trigger_name))

            longest_index_notation = max(
                longest_index_notation,
                helper.get_int_len(display) + helper.get_int_len(trigger_index)
            )

        longest_trigger_name += 3
        for display, trigger_index in enumerate(display_order):
            trigger = triggers[trigger_index]
            trigger_name = trigger.name

            name_buffer = longest_trigger_name - len(trigger_name)
            index_buffer = longest_index_notation - (helper.get_int_len(display) + helper.get_int_len(trigger_index))
            return_string += "\t" + trigger_name + (" " * name_buffer)
            return_string += f" [Index: {trigger_index}, Display: {display}] {' ' * index_buffer}"

            return_string += "\t(conditions: " + str(len(trigger.conditions)) + ", "
            return_string += " effects: " + str(len(trigger.effects)) + ")\n"

        variables = self.variables

        return_string += "\nVariables Summary:\n"
        if len(variables) == 0:
            return_string += "\t<< No Variables >>"

        longest_variable_name = -1
        for variable in variables:
            longest_variable_name = max(longest_variable_name, len(variable.name))

        longest_variable_name += 3
        for index, variable in enumerate(variables):
            var_name = variable.name
            name_buffer = " " * (longest_variable_name - len(var_name))
            return_string += f"\t{var_name}{name_buffer}[Index: {variable.variable_id}]\n"

        return return_string

    def get_content_as_string(self) -> str:
        return_string = "\nTriggers:\n"

        if len(self.triggers) == 0:
            return_string += "\t<<No triggers>>\n"

        for trigger_index in self.trigger_display_order:
            return_string += self.get_trigger_as_string(TS.index(trigger_index)) + "\n"

        return_string += "Variables:\n"

        if len(self.variables) == 0:
            return_string += "\t<<No Variables>>\n"

        for variable in self.variables:
            return_string += f"\t'{variable.name}' [Index: {variable.variable_id}]\n"

        return return_string

    def get_trigger_as_string(self, trigger_select: TriggerSelect) -> str:
        trigger_index, display_index, trigger = self._validate_and_retrieve_trigger_info(trigger_select)

        return_string = "\t'" + trigger.name + "'"
        return_string += " [Index: " + str(trigger_index) + ", Display: " + str(display_index) + "]" + ":\n"

        return_string += trigger.get_content_as_string()

        return return_string

    @staticmethod
    def _find_alterable_ce(trigger, trigger_ce_lock) -> (List[int], List[int]):
        lock_conditions = trigger_ce_lock.lock_conditions if trigger_ce_lock is not None else False
        lock_effects = trigger_ce_lock.lock_effects if trigger_ce_lock is not None else False
        lock_condition_type = trigger_ce_lock.lock_condition_type if trigger_ce_lock is not None else []
        lock_effect_type = trigger_ce_lock.lock_effect_type if trigger_ce_lock is not None else []
        lock_condition_ids = trigger_ce_lock.lock_condition_ids if trigger_ce_lock is not None else []
        lock_effect_ids = trigger_ce_lock.lock_effect_ids if trigger_ce_lock is not None else []

        alter_conditions: List[int] = []
        alter_effects: List[int] = []
        if not lock_conditions:
            for i, cond in enumerate(trigger.conditions):
                if i not in lock_condition_ids and cond.condition_type not in lock_condition_type:
                    alter_conditions.append(i)
        if not lock_effects:
            for i, effect in enumerate(trigger.effects):
                if i not in lock_effect_ids and effect.effect_type not in lock_effect_type:
                    alter_effects.append(i)

        return alter_conditions, alter_effects

    @staticmethod
    def _find_trigger_tree_nodes(trigger: TriggerObject) -> List[int]:
        return [
            effect.trigger_id for effect in trigger.effects if
            effect.effect_type in [Effect.ACTIVATE_TRIGGER, Effect.DEACTIVATE_TRIGGER]
        ]


class TriggerSelect:
    def __init__(self, trigger_index=None, display_index=None, trigger=None):
        """
        Object used to select a trigger in many trigger related functions. For ease of use, the alias `TS` can be
        called. You can also use those in combination with the class methods (factory methods). Like so:

        TS.index(4)    # To select the trigger with index 4

        TS.display(4)  # Trigger with display index 4

        TS.trigger(trigger)  # Well... The trigger object given...

        Args:
            trigger_index (int): The index of the trigger. Starting from 0, based on creation time
            display_index (int): The display index of a trigger. Starting from 0, based on display order in the editor
            trigger (TriggerObject): The trigger object itself.
        """
        self.trigger = trigger
        self.display_index = display_index
        self.trigger_index = trigger_index

    @classmethod
    def index(cls, index: int):
        return cls(trigger_index=index)

    @classmethod
    def display(cls, display_index: int):
        return cls(display_index=display_index)

    @classmethod
    def trigger(cls, trigger: TriggerObject):
        return cls(trigger=trigger)


TS = TriggerSelect


class TriggerCELock:
    def __init__(self, lock_conditions=False, lock_effects=False, lock_condition_type=None, lock_effect_type=None,
                 lock_condition_ids=None, lock_effect_ids=None):
        """
        Object used to identify which conditions and effects should be locked from change.

        Args:
            lock_conditions (bool): Lock all conditions
            lock_effects (bool): Lock all effects
            lock_condition_type (List[int]): Lock certain condition types. Example: `Condition.OWN_OBJECTS`
            lock_effect_type (List[int]): Lock certain effect types. Example: `Effect.CREATE_OBJECT`
            lock_condition_ids (List[int]): Lock certain conditions by their id
            lock_effect_ids (List[int]): Lock certain effects by their id
        """
        if lock_condition_type is None:
            lock_condition_type = []
        if lock_effect_type is None:
            lock_effect_type = []
        if lock_condition_ids is None:
            lock_condition_ids = []
        if lock_effect_ids is None:
            lock_effect_ids = []

        self.lock_conditions = lock_conditions
        self.lock_effects = lock_effects
        self.lock_condition_type = lock_condition_type
        self.lock_effect_type = lock_effect_type
        self.lock_condition_ids = lock_condition_ids
        self.lock_effect_ids = lock_effect_ids


class GroupBy(IntEnum):
    """
    Enum used to select what way the triggers should be grouped when using the function 'copy_trigger_tree_per_player'.
    Each mode has it's own explanation explained in the corresponding docstring.
    """
    NONE = -1
    """
    Don't group triggers. Add triggers to the end of the list. This could mean the source triggers aren't near the 
    copies. All copies, however, will be grouped like the TRIGGER grouping mechanism - at the end of the list.
    """
    TRIGGER = 0
    """
    Group triggers by trigger. All copies of a trigger will be put directly below the 'source' trigger. 
    Example: If the 'source' trigger has display_index of 3, and another trigger in it's tree has 5. Assuming the 
    triggers were copied for all players, display indexes of the copies from the first trigger would be 4-10. The second
    trigger in the tree would be bumped up to 12 (from 5) and it's copies would use: 13-19. 
    """
    PLAYER = 1
    """
    Group triggers by player. Warning: This mode could change some of the intended order of the triggers. All copies of
    the trigger will be put together sorted on player first, then trigger order. 
    Example: If the 'source' trigger has display_index of 3, and another trigger in it's tree has 5. Assuming the 
    triggers were copied for all players, display indexes of the copies from the firs trigger would be: 5, 7, 9, 11, 13,
    15 and 17. The second trigger would have moved to 4 (from 5) and it's copies would occupy slots: 6, 8, 10, 12, 14, 
    16 and 18. This way both Player.ONE (assuming that's the source player) triggers will be together, same goes for all
    other player. 
    
    In the example above can be seen how the second trigger was moved from 5 to 4. The unrelated trigger in between was
    moved to the end, 19. This might be an unwanted side effect, be aware.
    """



import struct

from AoE2ScenarioParser.helper import helper


def bytes_to_fixed_chars(byte_elements, codec="utf-8"):
    return byte_elements.decode(codec)


def fixed_chars_to_bytes(string, codec="utf-8"):
    return string.encode(codec)


_no_string_trail = [
    "filename",  # Scenario filename
    "ascii_instructions",  # Messages
    "ascii_hints",  # Messages
    "ascii_victory",  # Messages
    "ascii_loss",  # Messages
    "ascii_history",  # Messages
    "ascii_scouts",  # Messages
    "ascii_pregame",  # Cinematics
    "ascii_victory",  # Cinematics
    "ascii_loss",  # Cinematics
    "ascii_filename",  # Cinematics
    "strings",  # in PlayerDataTwoPiece
    "ai_names",  # in PlayerDataTwoPiece
    "ai_per_file_text",  # in AIStruct
    "unknown_string",  # in Map
    "map_color_mood",  # in Map
    "script_name",  # in Map
]


def bytes_to_str(byte_elements, codec="utf-8"):
    trail_removed_elements = helper.del_str_trail(byte_elements)
    try:
        return trail_removed_elements.decode(codec)
    except UnicodeDecodeError:
        return trail_removed_elements.decode('latin-1')


def str_to_bytes(string, retriever, codec="utf-8"):
    if retriever.name in _no_string_trail:
        return string.encode(codec)
    return helper.add_str_trail(string).encode(codec)


def bytes_to_int(byte_elements, endian="little", signed=False):
    return int.from_bytes(byte_elements, endian, signed=signed)


def int_to_bytes(integer: int, length, endian="little", signed=True):
    return integer.to_bytes(length, byteorder=endian, signed=signed)


def bytes_to_float(byte_elements):
    return struct.unpack('f', byte_elements)[0]


def float_to_bytes(f):
    return struct.pack('f', f)


def bytes_to_double(byte_elements):
    return struct.unpack('d', byte_elements)[0]


def double_to_bytes(d):
    return struct.pack('d', d)



from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.pieces.structs.aoe2_struct import AoE2Struct


class PlayerDataOneStruct(AoE2Struct):
    def __init__(self, parser_obj=None, data=None, pieces=None):
        retrievers = [
            Retriever("active", DataType("u32")),
            Retriever("human", DataType("u32")),
            Retriever("civilization", DataType("u32")),
            Retriever("architecture_set", DataType("u32")),
            Retriever("cty_mode", DataType("u32")),
        ]

        super().__init__("Player Data #1", retrievers, parser_obj, data, pieces=pieces)

    @staticmethod
    def defaults(pieces):
        # This default is adjusted in DataHeaderPiece:
        # - P1 (Active: 1, Human: 1)
        # - P2 (Active: 1)
        defaults = {
            'active': 0,
            'human': 0,
            'civilization': 36,
            'architecture_set': 36,
            'cty_mode': 4,
        }
        return defaults



from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.pieces import aoe2_piece


class MessagesPiece(aoe2_piece.AoE2Piece):
    def __init__(self, parser_obj=None, data=None, pieces=None):
        retrievers = [
            Retriever("instructions", DataType("u32")),
            Retriever("hints", DataType("u32")),
            Retriever("victory", DataType("u32")),
            Retriever("loss", DataType("u32")),
            Retriever("history", DataType("u32")),
            Retriever("scouts", DataType("u32")),
            Retriever("ascii_instructions", DataType("str16")),
            Retriever("ascii_hints", DataType("str16")),
            Retriever("ascii_victory", DataType("str16")),
            Retriever("ascii_loss", DataType("str16")),
            Retriever("ascii_history", DataType("str16")),
            Retriever("ascii_scouts", DataType("str16")),
        ]

        super().__init__("Messages", retrievers, parser_obj, data=data, pieces=pieces)

    @staticmethod
    def defaults(pieces):
        defaults = {
            'instructions': 4294967294,
            'hints': 4294967294,
            'victory': 4294967294,
            'loss': 4294967294,
            'history': 4294967294,
            'scouts': 4294967294,
            'ascii_instructions': '',
            'ascii_hints': '',
            'ascii_victory': 'This scenario was created using AoE2ScenarioParser! Hopefully you enjoyed!',
            'ascii_loss': '',
            'ascii_history': '',
            'ascii_scouts': '',
        }
        return defaults



from typing import Dict

from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.helper.retriever_dependency import RetrieverDependency
from AoE2ScenarioParser.pieces.structs.aoe2_struct import AoE2Struct


class ConditionStruct(AoE2Struct):
    dependencies: Dict[str, Dict[str, RetrieverDependency]] = {
        # "unit_ai_action": {
        #     "on_refresh": RetrieverDependency(
        #         DependencyAction.SET_REPEAT,
        #         DependencyTarget('FileHeaderPiece', 'version'),
        #         DependencyEval('1 if x in [\'1.40\'] else 0')
        #     ),
        #     "on_construct": RetrieverDependency(DependencyAction.REFRESH_SELF)
        # },
        # "unknown_4": {
        #     "on_refresh": RetrieverDependency(
        #         DependencyAction.SET_REPEAT,
        #         DependencyTarget('FileHeaderPiece', 'version'),
        #         DependencyEval('1 if x in [\'1.40\'] else 0')
        #     ),
        #     "on_construct": RetrieverDependency(DependencyAction.REFRESH_SELF)
        # },
        # "xs_function": {
        #     "on_refresh": RetrieverDependency(
        #         DependencyAction.SET_REPEAT,
        #         DependencyTarget('FileHeaderPiece', 'version'),
        #         DependencyEval('1 if x in [\'1.40\'] else 0')
        #     ),
        #     "on_construct": RetrieverDependency(DependencyAction.REFRESH_SELF)
        # },
    }

    def __init__(self, parser_obj=None, data=None, pieces=None):
        retrievers = [
            Retriever("condition_type", DataType("s32")),
            Retriever("static_value_21", DataType("s32")),  # Was always: 0x10
            Retriever("amount_or_quantity", DataType("s32")),  # Also technology state, also Difficulty
            Retriever("resource_type_or_tribute_list", DataType("s32")),
            Retriever("unit_object", DataType("s32")),
            Retriever("next_object", DataType("s32")),  # Old: Unit ID
            Retriever("object_list", DataType("s32")),
            Retriever("source_player", DataType("s32")),
            Retriever("technology", DataType("s32")),
            Retriever("timer", DataType("s32")),
            Retriever("unknown", DataType("s32")),
            Retriever("area_1_x", DataType("s32")),
            Retriever("area_1_y", DataType("s32")),
            Retriever("area_2_x", DataType("s32")),
            Retriever("area_2_y", DataType("s32")),
            Retriever("object_group", DataType("s32")),
            Retriever("object_type", DataType("s32")),  # -1: None, 1: Other, 2: Building, 3: Civilian, 4: Military
            Retriever("ai_signal", DataType("s32")),
            Retriever("inverted", DataType("s32")),
            Retriever("unknown_2", DataType("s32")),
            Retriever("variable", DataType("s32")),  # Number == VariableX
            Retriever("comparison", DataType("s32")),  # 0: ==, 1: <, 2: >, 3: <=, 4 >=
            Retriever("target_player", DataType("s32")),
            # Todo: Mark as 1.40 feature
            Retriever("unit_ai_action", DataType("s32")),
            # Todo: Mark as 1.40 feature
            Retriever("unknown_4", DataType("s32")),
            # Todo: Mark as 1.40 feature
            Retriever("xs_function", DataType("str32")),
        ]

        super().__init__("Condition", retrievers, parser_obj, data, pieces=pieces)

    @staticmethod
    def defaults(pieces):
        defaults = {
            'condition_type': 0,
            'static_value_21': 23,
            'amount_or_quantity': -1,
            'resource_type_or_tribute_list': -1,
            'unit_object': -1,
            'next_object': -1,
            'object_list': -1,
            'source_player': -1,
            'technology': -1,
            'timer': -1,
            'unknown': -1,
            'area_1_x': -1,
            'area_1_y': -1,
            'area_2_x': -1,
            'area_2_y': -1,
            'object_group': -1,
            'object_type': -1,
            'ai_signal': -1,
            'inverted': -1,
            'unknown_2': -1,
            'variable': -1,
            'comparison': -1,
            'target_player': -1,
            'unit_ai_action': -1,
            'unknown_4': -1,
            'xs_function': "",
        }
        return defaults



from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.pieces import aoe2_piece


class GlobalVictoryPiece(aoe2_piece.AoE2Piece):
    def __init__(self, parser_obj=None, data=None, pieces=None):
        retrievers = [
            Retriever("separator", DataType("u32")),
            Retriever("conquest_required", DataType("u32")),
            Retriever("ruins", DataType("u32")),
            Retriever("artifacts", DataType("u32")),
            Retriever("discovery", DataType("u32")),
            Retriever("explored_percent_of_map_required", DataType("u32")),
            Retriever("gold", DataType("u32")),
            Retriever("all_custom_conditions_required", DataType("u32")),
            Retriever("mode", DataType("u32")),
            Retriever("required_score_for_score_victory", DataType("u32")),
            Retriever("time_for_timed_game_in_10ths_of_a_year", DataType("u32")),
        ]

        super().__init__("Global Victory", retrievers, parser_obj, data=data, pieces=pieces)

    @staticmethod
    def defaults(pieces):
        defaults = {
            'separator': 4294967197,
            'conquest_required': 1,
            'ruins': 0,
            'artifacts': 0,
            'discovery': 0,
            'explored_percent_of_map_required': 0,
            'gold': 0,
            'all_custom_conditions_required': 0,
            'mode': 0,
            'required_score_for_score_victory': 900,
            'time_for_timed_game_in_10ths_of_a_year': 9000,
        }
        return defaults



from typing import Dict

from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.helper.retriever_dependency import RetrieverDependency, DependencyAction, DependencyTarget, \
    DependencyEval
from AoE2ScenarioParser.pieces.structs.aoe2_struct import AoE2Struct
from AoE2ScenarioParser.pieces.structs.unit import UnitStruct


class PlayerUnitsStruct(AoE2Struct):
    dependencies: Dict[str, Dict[str, RetrieverDependency]] = {
        "unit_count": {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_VALUE, DependencyTarget("self", "units"),
                DependencyEval("len(x)"))
        },
        "units": {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_REPEAT, DependencyTarget("self", "unit_count")),
            "on_construct": RetrieverDependency(DependencyAction.REFRESH_SELF),
            "on_commit": RetrieverDependency(
                DependencyAction.REFRESH, DependencyTarget("self", "unit_count"))
        },
    }

    def __init__(self, parser_obj=None, data=None, pieces=None):
        retrievers = [
            Retriever("unit_count", DataType("u32")),
            Retriever("units", DataType(UnitStruct))
        ]

        super().__init__("Player Units", retrievers, parser_obj, data, pieces=pieces)

    @staticmethod
    def defaults(pieces):
        defaults = {
            'unit_count': 0,
            'units': [],
        }
        return defaults



from AoE2ScenarioParser.objects.aoe2_object import AoE2Object


class DataHeaderObject(AoE2Object):
    def __init__(self,
                 version,
                 filename,
                 ):
        self.version = version
        self.filename = filename

        super().__init__()



import collections
import time
import zlib
from collections import OrderedDict
from typing import List, Type

from AoE2ScenarioParser.helper import generator, helper
from AoE2ScenarioParser.helper import parser
from AoE2ScenarioParser.helper.helper import create_textual_hex, SimpleLogger
from AoE2ScenarioParser.helper.retriever import get_retriever_by_name
from AoE2ScenarioParser.objects.aoe2_object_manager import AoE2ObjectManager
from AoE2ScenarioParser.objects.map_obj import MapObject
from AoE2ScenarioParser.objects.triggers_obj import TriggersObject
from AoE2ScenarioParser.objects.units_obj import UnitsObject
from AoE2ScenarioParser.pieces.aoe2_piece import AoE2Piece
from AoE2ScenarioParser.pieces.background_image import BackgroundImagePiece
from AoE2ScenarioParser.pieces.cinematics import CinematicsPiece
from AoE2ScenarioParser.pieces.data_header import DataHeaderPiece
from AoE2ScenarioParser.pieces.diplomacy import DiplomacyPiece
from AoE2ScenarioParser.pieces.file_header import FileHeaderPiece
from AoE2ScenarioParser.pieces.files import FilesPiece
from AoE2ScenarioParser.pieces.global_victory import GlobalVictoryPiece
from AoE2ScenarioParser.pieces.map import MapPiece
from AoE2ScenarioParser.pieces.messages import MessagesPiece
from AoE2ScenarioParser.pieces.options import OptionsPiece
from AoE2ScenarioParser.pieces.player_data_two import PlayerDataTwoPiece
from AoE2ScenarioParser.pieces.triggers import TriggerPiece
from AoE2ScenarioParser.pieces.units import UnitsPiece


class AoE2Scenario:
    @property
    def trigger_manager(self) -> TriggersObject:
        return self._object_manager.objects['TriggersObject']

    @property
    def unit_manager(self) -> UnitsObject:
        return self._object_manager.objects['UnitsObject']

    @property
    def map_manager(self) -> MapObject:
        return self._object_manager.objects['MapObject']

    def __init__(self):
        self.read_mode = None
        self.parser = None
        self._file_header = None
        self._decompressed_file_data = None
        self._file = None
        self._object_manager = None

    @classmethod
    def from_file(cls, filename, log_reading=True, log_parsing=True):
        scenario = cls()
        scenario.read_mode = "from_file"

        print("\nPreparing & Loading file: '" + filename + "'...")
        scenario_file = open(filename, "rb")
        scenario._file = scenario_file.read()
        scenario_file.seek(0)  # Reset file cursor to 0

        scenario._file_header = scenario_file.read(scenario._compute_header_length())
        scenario._decompressed_file_data = zlib.decompress(scenario_file.read(), -zlib.MAX_WBITS)

        scenario_file.close()
        print("File prepared and loaded.")

        scenario.parser = parser.Parser()
        scenario._read_file(log_reading=log_reading)
        scenario._object_manager = AoE2ObjectManager(scenario._parsed_header, scenario._parsed_data,
                                                     log_parsing=log_parsing)
        return scenario

    @classmethod
    def create_default(cls, log_creating=True, log_parsing=False):
        scenario = cls()
        scenario.read_mode = "create_default"

        lgr = SimpleLogger(log_creating)
        lgr.print("\nFile creation started...")

        scenario.parser = parser.Parser()
        scenario._parsed_header = collections.OrderedDict()
        scenario._parsed_data = collections.OrderedDict()
        pieces = OrderedDict(**scenario._parsed_header, **scenario._parsed_data)

        for piece in _header_structure:
            piece_name = piece.__name__
            lgr.print("\tCreating " + piece_name + "...", replace_line=True)
            scenario._parsed_header[piece_name] = piece(scenario.parser, data=list(piece.defaults(pieces).values()),
                                                        pieces=pieces)
            lgr.print("\tCreating " + piece_name + " finished successfully.", replace_line=True)
            lgr.print()
        for piece in _file_structure:
            pieces = OrderedDict(**scenario._parsed_header, **scenario._parsed_data)
            piece_name = piece.__name__
            lgr.print("\tCreating " + piece_name + "...", replace_line=True)
            scenario._parsed_data[piece_name] = piece(scenario.parser, data=list(piece.defaults(pieces).values()),
                                                      pieces=pieces)
            lgr.print("\tCreating " + piece_name + " finished successfully.", replace_line=True)
            lgr.print()
        lgr.print("File creation finished successfully.")

        scenario._object_manager = AoE2ObjectManager(scenario._parsed_header, scenario._parsed_data,
                                                     log_parsing=log_parsing)
        return scenario

    def _read_file(self, log_reading):
        lgr = SimpleLogger(log_reading)
        lgr.print("\nFile reading started...")

        self._parsed_header = collections.OrderedDict()
        self._parsed_data = collections.OrderedDict()
        header_generator = self._create_header_generator(1)
        data_generator = self._create_data_generator(1)

        current_piece = ""
        try:
            for piece_object in _header_structure:
                # Rerender pieces dict each time - changes constantly
                pieces = collections.OrderedDict(**self._parsed_header, **self._parsed_data)
                piece = piece_object(self.parser)
                piece_name = type(piece).__name__
                self._parsed_header[piece_name] = piece
                current_piece = piece_name

                lgr.print("\tReading " + piece_name + "...", replace_line=True)
                piece.set_data_from_generator(header_generator, pieces)
                lgr.print("\tReading " + piece_name + " finished successfully.", replace_line=True)
                lgr.print()

            for piece_object in _file_structure:
                # Rerender pieces dict each time - changes constantly
                pieces = collections.OrderedDict(**self._parsed_header, **self._parsed_data)
                piece = piece_object(self.parser)
                piece_name = type(piece).__name__
                self._parsed_data[piece_name] = piece
                current_piece = piece_name

                lgr.print("\tReading " + piece_name + "...", replace_line=True)
                piece.set_data_from_generator(data_generator, pieces)
                lgr.print("\tReading " + piece_name + " finished successfully.", replace_line=True)
                lgr.print()
        except Exception as e:
            print(f"\n[{e.__class__.__name__}] [EXIT] AoE2Scenario._read_file: \n\tPiece: {current_piece}\n")
            print("Writing ErrorFile...")
            self._debug_byte_structure_to_file(
                filename="../ErrorFile",
                generator_for_trail=data_generator,
                log_debug_write=True
            )
            time.sleep(1)
            print("\nErrorFile written. \n\n\n ------------------------ STACK TRACE ------------------------\n\n")
            time.sleep(1)
            raise e
        lgr.print("File reading finished successfully.")

    def write_to_file(self, filename, commit_on_write=True, log_writing=True, log_reconstructing=True):
        self._write_from_structure(
            filename,
            log_writing=log_writing,
            log_reconstructing=log_reconstructing,
            commit_on_write=commit_on_write,
        )

    def _write_from_structure(self,
                              filename,
                              write_in_bytes=True,
                              compress=True,
                              commit_on_write=True,
                              log_writing=True,
                              log_reconstructing=True):
        if hasattr(self, '_object_manager') and commit_on_write:
            self._object_manager.reconstruct(log_reconstructing=log_reconstructing)
        lgr = SimpleLogger(log_writing)
        lgr.print("\nFile writing from structure started...")

        pieces = collections.OrderedDict(**self._parsed_header, **self._parsed_data)

        byte_header_list = []
        byte_data_list = []
        for key in self._parsed_header:
            lgr.print("\twriting " + key + "...", replace_line=True)
            for retriever in self._parsed_header[key].retrievers:
                byte_header_list.append(parser.retriever_to_bytes(retriever, pieces))
            lgr.print("\twriting " + key + " finished successfully.", replace_line=True)
            lgr.print()

        for key in self._parsed_data:
            lgr.print("\twriting " + key + "...", replace_line=True)
            for retriever in self._parsed_data[key].retrievers:
                try:
                    byte_data_list.append(parser.retriever_to_bytes(retriever, pieces))
                except AttributeError as e:
                    print("AttributeError occurred while writing '" + key + "' > '" + retriever.name + "'")
                    print("\n\n\nAn error occurred. Writing failed.")
                    raise e
            lgr.print("\twriting " + key + " finished successfully.", replace_line=True)
            lgr.print()

        file = open(filename, "wb" if write_in_bytes else "w")

        byte_header = b''.join(byte_header_list)
        byte_data = b''.join(byte_data_list)

        file.write(byte_header if write_in_bytes else create_textual_hex(byte_header.hex()))
        if compress:
            lgr.print("\tCompressing...", replace_line=True)
            # https://stackoverflow.com/questions/3122145/zlib-error-error-3-while-decompressing-incorrect-header-check/22310760#22310760
            deflate_obj = zlib.compressobj(9, zlib.DEFLATED, -zlib.MAX_WBITS)
            compressed = deflate_obj.compress(b''.join(byte_data_list)) + deflate_obj.flush()
            file.write(compressed if write_in_bytes else create_textual_hex(compressed.hex()))
            lgr.print("\tCompressing finished successfully.", replace_line=True)
            lgr.print()
        else:
            file.write(byte_data if write_in_bytes else create_textual_hex(byte_data.hex()))

        file.close()
        lgr.print("File writing finished successfully.")

    def _create_header_generator(self, chunk_size):
        return generator.create_generator(self._file_header, chunk_size)

    def _create_data_generator(self, chunk_size):
        return generator.create_generator(self._decompressed_file_data, chunk_size)

    def _create_file_generator(self, chunk_size):
        return generator.create_generator(self._file, chunk_size)

    def _compute_header_length(self):
        return parser.calculate_length(
            self._create_file_generator(1),
            FileHeaderPiece(parser.Parser()).retrievers
        )

    """ #############################################
    ################ Debug functions ################
    ############################################# """

    def _debug_write_from_source(self, filename, datatype, write_bytes=True):
        """This function is used as a test debugging writing. It writes parts of the read file to the filesystem."""
        print("File writing from source started with attributes " + datatype + "...")
        file = open(filename, "wb" if write_bytes else "w")
        selected_parts = []
        for t in datatype:
            if t == "f":
                selected_parts.append(self._file)
            elif t == "h":
                selected_parts.append(self._file_header)
            elif t == "d":
                selected_parts.append(self._decompressed_file_data)
        parts = None
        for part in selected_parts:
            if parts is None:
                parts = part
                continue
            parts += part
        file.write(parts if write_bytes else create_textual_hex(parts.hex()))
        file.close()
        print("File writing finished successfully.")

    def _debug_byte_structure_to_file(self, filename, generator_for_trail=None, log_debug_write=True, commit=False):
        """ Used for debugging - Writes structure from read file to the filesystem in a easily readable manner. """
        if commit and hasattr(self, '_object_manager'):
            # self._object_manager.reconstruct(log_debug_write)
            self._write_from_structure(filename, log_writing=log_debug_write, log_reconstructing=log_debug_write)

        lgr = SimpleLogger(log_debug_write)

        pieces = collections.OrderedDict(**self._parsed_header, **self._parsed_data)
        lgr.print("\nWriting structure to file...")
        with open(filename, 'w', encoding="utf-8") as output_file:
            result = []
            for key in self._parsed_header:
                lgr.print("\tWriting " + key + "...", replace_line=True)
                result.append(self._parsed_header[key].get_byte_structure_as_string(pieces))
                lgr.print("\tWriting " + key + " finished successfully.", replace_line=True)
                lgr.print()
            for key in self._parsed_data:
                lgr.print("\tWriting " + key + "...", replace_line=True)
                result.append(self._parsed_data[key].get_byte_structure_as_string(pieces))
                lgr.print("\tWriting " + key + " finished successfully.", replace_line=True)
                lgr.print()

            if generator_for_trail is not None:
                lgr.print("\tWriting trail...", replace_line=True)
                trail_length = -1  # -1 == inf
                try:
                    trail = b''
                    i = 0
                    while i != trail_length:
                        trail += generator.repeat_generator(
                            generator=generator_for_trail,
                            run_times=1,
                            intended_stop_iteration=True,
                            return_bytes=True
                        )
                        i += 1
                except StopIteration:
                    pass  # Expected, if trail is not present or shorter than {trail_length} bytes
                if i != 0:
                    i += 1
                if i == trail_length:
                    i = str(i) + '+'
                if trail_length == -1:
                    trail_length = i
                result.append(f"\n\n{'#' * 27} TRAIL ({i}/{trail_length})\n\n")
                result.append(helper.create_textual_hex(trail.hex(), space_distance=2, enter_distance=24))
                lgr.print("\tWriting trail finished successfully.", replace_line=True)
                lgr.print()

            output_file.write(''.join(result))
            output_file.close()
        lgr.print("Writing structure to file finished successfully.")


_header_structure: List[Type[AoE2Piece]] = [
    FileHeaderPiece
]
_file_structure: List[Type[AoE2Piece]] = [
    DataHeaderPiece,
    MessagesPiece,
    CinematicsPiece,
    BackgroundImagePiece,
    PlayerDataTwoPiece,
    GlobalVictoryPiece,
    DiplomacyPiece,
    OptionsPiece,
    MapPiece,
    UnitsPiece,
    TriggerPiece,
    FilesPiece,
]

# Define piece names
data_header_piece = "DataHeaderPiece"
messages_piece = "MessagesPiece"
cinematics_piece = "CinematicsPiece"
background_image_piece = "BackgroundImagePiece"
player_data_two_piece = "PlayerDataTwoPiece"
global_victory_piece = "GlobalVictoryPiece"
diplomacy_piece = "DiplomacyPiece"
options_piece = "OptionsPiece"
map_piece = "MapPiece"
units_piece = "UnitsPiece"
trigger_piece = "TriggerPiece"
files_piece = "FilesPiece"



from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.pieces.structs.aoe2_struct import AoE2Struct


class VariableStruct(AoE2Struct):
    def __init__(self, parser_obj=None, data=None, pieces=None):
        retrievers = [
            Retriever("variable_id", DataType("u32")),
            Retriever("name", DataType("str32")),
        ]

        super().__init__("Variable", retrievers, parser_obj, data, pieces=pieces)

    @staticmethod
    def defaults(pieces):
        defaults = {
            'variable_id': 0,
            'name': '_Variable0',
        }
        return defaults



import math
import sys
from enum import IntEnum
from typing import List, Dict

from AoE2ScenarioParser.datasets.buildings import Building, GaiaBuilding
from AoE2ScenarioParser.datasets.heroes import Hero
from AoE2ScenarioParser.datasets.units import Unit, GaiaUnit

""" =============================================================
========================= HEX FUNCTIONS =========================
=============================================================="""


def update_order_array(order_array, supposed_length):
    for i in range(0, supposed_length):
        if i not in order_array:
            order_array.append(i)


def create_textual_hex(string, space_distance=2, enter_distance=48):
    """Please note that the 'enter_distance' parameter is including the - to be added - spaces. If you calculated it
    without the spaces, please multiply the number by: `block size incl space / block size excl space`"""
    return insert_char(insert_char(string, " ", space_distance), "\n", enter_distance)


# Credits: gurney alex @ https://stackoverflow.com/a/2657733/7230293
def insert_char(string, char, step=64):
    return char.join(string[i:i + step] for i in range(0, len(string), step))


""" =============================================================
======================= STRING MODIFIERS ========================
=============================================================="""


def add_str_trail(string):
    if len(string) > 0:
        string = string + ("\x00" if string[-1] != "\x00" else "")
    return string


def del_str_trail(string):
    if len(string) > 0:
        string = string if string[-1] != "\x00" and string[-1] != 0 else string[:-1]
    return string


def add_prefix_chars(string, char, length):
    if len(string) > length:
        return string
    else:
        return char * (length - len(string)) + string


def add_suffix_chars(string, char, total_length):
    if len(string) > total_length:
        return string
    else:
        return string + char * (total_length - len(string))


def q_str(value: any) -> str:
    if type(value) is str:
        return f"'{value}'"
    return str(value)


""" =============================================================
======================== PRETTY PRINTERS ========================
=============================================================="""

_default_inline_types = {
    'int': 8,
    'float': 8
}


def pretty_print_list(plist: List, inline_types: Dict[str, int] = None):
    if len(plist) == 0:
        return "[]\r\n"
    if inline_types is None:
        inline_types = _default_inline_types
    entry_type = type(plist[0]).__name__  # Get entry type

    return_string = "[\r\n"
    line_items = []
    for index, entry in enumerate(plist):
        if entry_type in inline_types.keys():
            line_items.append(entry)
            if index % inline_types[entry_type] == inline_types[entry_type] - 1:
                return_string += _create_inline_line(line_items)
                line_items = []
            continue
        else:
            return_string += f"\t{entry}\r\n"
    if len(line_items) != 0:
        return_string += _create_inline_line(line_items)
    return return_string + "]\r\n"


def _create_inline_line(entries):
    return "\t" + ", ".join(map(str, entries)) + "\r\n"


def pretty_print_dict(pdict: dict):
    return_string = "{\n"
    for key, value in pdict.items():
        newline = f"\t{key}: {value}"
        if newline[::-2] != "\n":
            newline += "\n"
        return_string += newline
    return return_string + "}\n"


def pretty_print_name(name: str) -> str:
    """
    Returns a pretty-printed version of the name string.
    Replaces all underscores with spaces and capitalizes the first letter
    of each word.
    For example, elite_chu_ko_nu -> Elite Chu Ko Nu.

    :Author:
        T-West (https://github.com/twestura/)
    """
    return ' '.join(s[0].upper() + s[1:] for s in name.split('_'))


""" =============================================================
============================ COORDS =============================
=============================================================="""


def xy_to_i(x, y, map_size):
    return x * map_size + y


def i_to_xy(i, map_size):
    return int(i / map_size), i % map_size


""" =============================================================
============================= HASH ==============================
=============================================================="""


def hash_list(lst: list):
    return hash(tuple(lst))


def list_changed(lst, lst_hash):
    return lst_hash != hash(tuple(lst))


""" =============================================================
============================= OTHER =============================
=============================================================="""


def get_enum_from_unit_const(const: int) -> IntEnum:
    """
    Returns an Enum corresponding with the given Const.

    Arguments:
        const: The constant representing a unit
    """
    if any(item == const for item in Unit):
        return Unit(const)
    if any(item.value == const for item in Building):
        return Building(const)
    if any(item.value == const for item in Hero):
        return Hero(const)
    if any(item == const for item in GaiaUnit):
        return GaiaUnit(const)
    if any(item == const for item in GaiaBuilding):
        return GaiaBuilding(const)


def get_int_len(num):
    if num > 0:
        return math.floor(math.log10(num))
    return 0


def evaluate_index_params(x_id, display_index, name):
    if x_id is None and display_index is None:
        raise ValueError(f"Please choose '{name}_id' or 'display_index' as identification for the wanted {name}")
    if x_id is not None and display_index is not None:
        raise ValueError(f"Please identify a {name} using '{name}_id' or 'display_index' but not both")


class SimpleLogger:
    def __init__(self, should_log):
        self.should_log = should_log

    def print(self, string="", end="\n", replace_line=False):
        if self.should_log:
            if not replace_line:
                print(string, end=end)
            else:
                sys.stdout.write('\r' + string)

    def __repr__(self):
        return f"SimpleLogger[should_log: {self.should_log}]"


class Tile:
    def __init__(self, x: int, y: int):
        self.x: float = x
        self.y: float = y

    def __repr__(self):
        return f"Tile[x: {self.x}, y: {self.y}]"

    @property
    def x(self) -> float:
        return self._x

    @x.setter
    def x(self, val: int):
        self._x = int(val) + .5

    @property
    def y(self) -> float:
        return self._y

    @y.setter
    def y(self, val: int):
        self._y = int(val) + .5

    @property
    def x1(self) -> int:
        return int(self.x - .5)

    @property
    def y1(self) -> int:
        return int(self.y - .5)

    @property
    def x2(self) -> int:
        return int(self.x + .5)

    @property
    def y2(self) -> int:
        return int(self.y + .5)



from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.pieces import aoe2_piece
from AoE2ScenarioParser.pieces.structs.player_diplomacy import PlayerDiplomacyStruct


class DiplomacyPiece(aoe2_piece.AoE2Piece):
    def __init__(self, parser_obj=None, data=None, pieces=None):
        retrievers = [
            Retriever("per_player_diplomacy", DataType(PlayerDiplomacyStruct, repeat=16)),
            Retriever("individual_victories", DataType("60", repeat=16 * 12)),  # 12 Conditions per (16) Player(s).
            Retriever("separator", DataType("u32")),
            Retriever("per_player_allied_victory", DataType("u32", repeat=16)),  # Ignored -> PlayerDataThree
            Retriever("unknown", DataType("4")),
        ]

        super().__init__("Diplomacy", retrievers, parser_obj, data=data, pieces=pieces)

    @staticmethod
    def defaults(pieces):
        defaults = {
            'per_player_diplomacy': [
                PlayerDiplomacyStruct(data=list(PlayerDiplomacyStruct.defaults(pieces).values()), pieces=pieces) for _
                in range(16)
            ],
            'individual_victories': [b'\x00' * 60] * 192,
            'separator': 4294967197,
            'per_player_allied_victory': [0] * 16,
            'unknown': b'\x00\x01\x00\x04',
        }
        return defaults



from __future__ import annotations

from typing import List, Union, TYPE_CHECKING

from AoE2ScenarioParser.helper import helper
from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever_dependency import RetrieverDependency

if TYPE_CHECKING:
    from AoE2ScenarioParser.helper.retriever_object_link import RetrieverObjectLink


class Retriever:
    on_construct: RetrieverDependency
    on_commit: RetrieverDependency
    on_refresh: RetrieverDependency
    """ A Class for defining how to retrieve data.
    The Constructor has quite some parameters which can all be used for getting the proper data
    """

    def __init__(self, name, datatype=DataType(), possibly_list=True, log_value=False):
        """
        Args:
            name (str): The name of the item. Has to be unique within the Piece or Struct
            datatype (DataType): A datatype object
            log_value (bool): A boolean for, mostly, debugging. This will log this Retriever with it's data when the
                data is changed, when this retriever is constructed and committed.
        """
        self.name: str = name
        self.datatype: DataType = datatype
        if log_value:
            self.datatype.log_value = True
        self.datatype._debug_retriever_name = name
        self.possibly_list = possibly_list
        self.log_value = log_value
        self._data = None

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, value):
        old_value = ""
        if hasattr(self, 'data'):
            old_value = self.data
        self._data = value
        if self.log_value:
            self._update_print(old_value, value)

    def _update_print(self, old, new):
        print(f"{self.to_simple_string()} >>> set to: {helper.q_str(new)} (was: {helper.q_str(old)})")

    def get_short_str(self):
        if self.data is not None:
            if type(self.data) is list:
                data = helper.pretty_print_list(self.data)
            else:
                data = helper.q_str(self.data)
            return self.name + " (" + self.datatype.to_simple_string() + "): " + data

    def to_simple_string(self):
        return f"[Retriever] {self.name}: {self.datatype}"

    def __repr__(self):
        if type(self.data) is list:
            data = str(helper.pretty_print_list(self.data))
        else:
            data = helper.q_str(self.data)
        return f"{self.to_simple_string()} >>> {data}"


def get_retriever_by_name(retriever_list: List[Union[Retriever, RetrieverObjectLink]], name: str) \
        -> Union[Retriever, RetrieverObjectLink]:
    for retriever in retriever_list:
        if retriever.name == name:
            return retriever



from typing import Dict

from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.helper.retriever_dependency import RetrieverDependency, DependencyAction, DependencyTarget, \
    DependencyEval
from AoE2ScenarioParser.pieces import aoe2_piece
from AoE2ScenarioParser.pieces.structs.trigger import TriggerStruct
from AoE2ScenarioParser.pieces.structs.variable import VariableStruct


class TriggerPiece(aoe2_piece.AoE2Piece):
    dependencies: Dict[str, Dict[str, RetrieverDependency]] = {
        "number_of_triggers": {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_VALUE, DependencyTarget("self", "trigger_data"),
                DependencyEval("len(x)")
            )
        },
        "trigger_data": {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_REPEAT, DependencyTarget("self", "number_of_triggers")
            ),
            "on_construct": RetrieverDependency(DependencyAction.REFRESH_SELF),
            "on_commit": RetrieverDependency(
                DependencyAction.REFRESH, DependencyTarget(
                    [
                        "self",
                        "FileHeaderPiece",
                        "OptionsPiece",
                    ], [
                        "number_of_triggers",
                        "trigger_count",
                        "number_of_triggers",
                    ]
                )
            )
        },
        "trigger_display_order_array": {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_REPEAT, DependencyTarget("self", "number_of_triggers")
            ),
            "on_construct": RetrieverDependency(DependencyAction.REFRESH_SELF)
        },
        "number_of_variables": {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_VALUE, DependencyTarget("self", "variable_data"),
                DependencyEval("len(x)")
            )
        },
        "variable_data": {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_REPEAT, DependencyTarget("self", "number_of_variables")
            ),
            "on_construct": RetrieverDependency(DependencyAction.REFRESH_SELF),
            "on_commit": RetrieverDependency(
                DependencyAction.REFRESH, DependencyTarget("self", "number_of_variables")
            )
        },
    }

    def __init__(self, parser_obj=None, data=None, pieces=None):
        retrievers = [
            Retriever("trigger_version", DataType("f64")),
            Retriever("trigger_instruction_start", DataType("s8")),
            Retriever("number_of_triggers", DataType("s32")),
            Retriever("trigger_data", DataType(TriggerStruct)),
            Retriever("trigger_display_order_array", DataType("u32")),
            Retriever("unknown", DataType("1028")),
            Retriever("number_of_variables", DataType("u32")),
            Retriever("variable_data", DataType(VariableStruct)),
        ]

        super().__init__("Triggers", retrievers, parser_obj, data=data, pieces=pieces)

    @staticmethod
    def defaults(pieces):
        defaults = {
            'trigger_version': 2.4,
            'trigger_instruction_start': 0,
            'number_of_triggers': 0,
            'trigger_data': [],
            'trigger_display_order_array': [],
            'unknown': b'\x00' * 1028,
            'number_of_variables': 0,
            'variable_data': [],
        }
        return defaults



from typing import Dict

from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.helper.retriever_dependency import RetrieverDependency, DependencyAction, DependencyTarget, \
    DependencyEval
from AoE2ScenarioParser.pieces import aoe2_piece
from AoE2ScenarioParser.pieces.structs.ai2 import AI2Struct


class FilesPiece(aoe2_piece.AoE2Piece):
    dependencies: Dict[str, Dict[str, RetrieverDependency]] = {
        "script_file_path": {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_VALUE, DependencyTarget("MapPiece", "script_name"),
                DependencyEval("x + ('.xs' if len(x) > 0 else '')")
            ),
        },
        "ai_files_present": {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_VALUE, DependencyTarget("self", "number_of_ai_files"),
                DependencyEval('0 if x == 0 else 1')
            ),
        },
        "number_of_ai_files": {
            "on_construct": RetrieverDependency(
                DependencyAction.SET_REPEAT, DependencyTarget("self", "ai_files_present"),
            ),
            "on_refresh": [
                # RetrieverDependency(
                #     DependencyAction.SET_VALUE, DependencyTarget("self", "ai_files"), DependencyEval('len(x)')
                # ),
                RetrieverDependency(
                    DependencyAction.SET_REPEAT, DependencyTarget("self", "ai_files"),
                    DependencyEval('1 if len(x) > 0 else 0')
                ),
            ],
            "on_commit": RetrieverDependency(
                DependencyAction.REFRESH, DependencyTarget("self", "ai_files_present")
            )
        },
        "ai_files": {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_REPEAT, DependencyTarget("self", "number_of_ai_files"),
                DependencyEval('0 if x == [] else x')  # x is a list because of the SET_REPEAT present
            ),
            "on_construct": RetrieverDependency(DependencyAction.REFRESH_SELF),
            "on_commit": RetrieverDependency(
                DependencyAction.REFRESH, DependencyTarget("self", "number_of_ai_files")
            )
        },
    }

    def __init__(self, parser_obj=None, data=None, pieces=None):
        retrievers = [
            Retriever("unknown_2", DataType("4")),
            Retriever("script_file_path", DataType("str16")),
            Retriever("script_file_content", DataType("str32")),
            Retriever("ai_files_present", DataType("u32")),
            Retriever("unknown_4", DataType("4")),
            Retriever("number_of_ai_files", DataType("u32"), possibly_list=False),
            Retriever("ai_files", DataType(AI2Struct)),
            Retriever("__END_OF_FILE_MARK__", DataType("1")),
        ]

        super().__init__("Files", retrievers, parser_obj, data=data, pieces=pieces)

    @staticmethod
    def defaults(pieces):
        defaults = {
            'unknown_2': b'\x00' * 4,
            'script_name': "",
            'script_file_content': "",
            'ai_files_present': 0,
            'unknown_4': b'\x00' * 4,
            'number_of_ai_files': [],
            'ai_files': [],
            '__END_OF_FILE_MARK__': b'',
        }
        return defaults



from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.helper.retriever_dependency import RetrieverDependency, DependencyAction, DependencyTarget, \
    DependencyEval
from AoE2ScenarioParser.pieces.structs.aoe2_struct import AoE2Struct


class BitMapInfoStruct(AoE2Struct):
    dependencies = {
        'colors_used': {
            "on_construct": RetrieverDependency(
                DependencyAction.SET_REPEAT, DependencyTarget("self", "number_of_colors_used"),
            )
        },
        'image': {
            "on_construct": RetrieverDependency(
                DependencyAction.SET_REPEAT, DependencyTarget(["self"] * 2, ["width", "height"]),
                DependencyEval("width * height", values_as_variable=['width', 'height'])
            )
        },
    }

    def __init__(self, parser_obj=None, data=None, pieces=None):
        retrievers = [
            Retriever("size", DataType("s32")),
            Retriever("width", DataType("u32")),
            Retriever("height", DataType("u32")),
            Retriever("planes", DataType("s16")),
            Retriever("bit_count", DataType("s16")),
            Retriever("compression", DataType("u32")),
            Retriever("image_size", DataType("u32")),
            Retriever("x_pels", DataType("u32")),
            Retriever("y_pels", DataType("u32")),
            Retriever("number_of_colors_used", DataType("u32")),
            Retriever("important_colors", DataType("u32")),
            Retriever("colors_used", DataType("u32")),
            Retriever("image", DataType("u8")),
        ]

        super().__init__("BitMap Info", retrievers, parser_obj, data=data, pieces=pieces)

    @staticmethod
    def defaults(pieces):
        defaults = {}
        return defaults



from typing import Dict

from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.helper.retriever_dependency import RetrieverDependency, DependencyAction, DependencyTarget, \
    DependencyEval
from AoE2ScenarioParser.pieces.structs.aoe2_struct import AoE2Struct


class EffectStruct(AoE2Struct):
    dependencies: Dict[str, Dict[str, RetrieverDependency]] = {
        "effect_type": {
            "on_commit": RetrieverDependency(
                DependencyAction.REFRESH,
                DependencyTarget(['self', 'self', 'self'], ['aa_quantity', 'aa_armor_or_attack_type', 'quantity'])
            )
        },
        "aa_quantity": {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_REPEAT,
                DependencyTarget('self', 'effect_type'),
                DependencyEval('1 if x in [28, 31] else 0')
            ),
            "on_construct": RetrieverDependency(DependencyAction.REFRESH_SELF)
        },
        "aa_armor_or_attack_type": {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_REPEAT,
                DependencyTarget('self', 'effect_type'),
                DependencyEval('1 if x in [28, 31] else 0')
            ),
            "on_construct": RetrieverDependency(DependencyAction.REFRESH_SELF)
        },
        "quantity": {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_REPEAT,
                DependencyTarget('self', 'effect_type'),
                DependencyEval('1 if x not in [28, 31] else 0')
            ),
            "on_construct": RetrieverDependency(DependencyAction.REFRESH_SELF)
        },
        "number_of_units_selected": {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_VALUE,
                DependencyTarget('self', 'selected_object_ids'),
                DependencyEval('len(x) if len(x) != 0 else -1')
            )
        },
        "selected_object_ids": {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_REPEAT,
                DependencyTarget("self", "number_of_units_selected"),
            ),
            "on_construct": RetrieverDependency(DependencyAction.REFRESH_SELF),
            "on_commit": RetrieverDependency(
                DependencyAction.REFRESH,
                DependencyTarget("self", "number_of_units_selected")
            )
        },
        # "player_color": {
        #     "on_refresh": RetrieverDependency(
        #         DependencyAction.SET_REPEAT,
        #         DependencyTarget('FileHeaderPiece', 'version'),
        #         DependencyEval('1 if x in [\'1.40\'] else 0')
        #     ),
        #     "on_construct": RetrieverDependency(DependencyAction.REFRESH_SELF)
        # },
    }

    def __init__(self, parser_obj=None, data=None, pieces=None):
        retrievers = [
            Retriever("effect_type", DataType("s32")),
            Retriever("static_value_46", DataType("s32")),  # always 0x17, now 0x30 (48)?
            Retriever("ai_script_goal", DataType("s32")),
            Retriever("aa_quantity", DataType("u8"), possibly_list=False),
            Retriever("aa_armor_or_attack_type", DataType("s24"), possibly_list=False),
            Retriever("quantity", DataType("s32"), possibly_list=False),
            Retriever("tribute_list", DataType("s32")),
            Retriever("diplomacy", DataType("s32")),
            Retriever("number_of_units_selected", DataType("s32")),
            Retriever("unknown", DataType("s32")),
            Retriever("object_list_unit_id", DataType("s32")),
            Retriever("source_player", DataType("s32")),
            Retriever("target_player", DataType("s32")),
            Retriever("technology", DataType("s32")),
            Retriever("string_id", DataType("s32")),
            Retriever("unknown_2", DataType("s32")),
            Retriever("display_time", DataType("s32")),
            Retriever("trigger_id", DataType("s32")),
            Retriever("location_x", DataType("s32")),
            Retriever("location_y", DataType("s32")),
            Retriever("area_1_x", DataType("s32")),
            Retriever("area_1_y", DataType("s32")),
            Retriever("area_2_x", DataType("s32")),
            Retriever("area_2_y", DataType("s32")),
            Retriever("object_group", DataType("s32")),
            Retriever("object_type", DataType("s32")),
            Retriever("instruction_panel_position", DataType("s32")),
            Retriever("attack_stance", DataType("s32")),
            Retriever("time_unit", DataType("s32")),
            Retriever("enabled_or_victory", DataType("s32")),
            Retriever("food", DataType("s32")),
            Retriever("wood", DataType("s32")),
            Retriever("stone", DataType("s32")),
            Retriever("gold", DataType("s32")),
            Retriever("item_id", DataType("s32")),
            Retriever("flash_object", DataType("s32")),
            Retriever("force_research_technology", DataType("s32")),
            Retriever("visibility_state", DataType("s32")),
            Retriever("scroll", DataType("s32")),
            Retriever("operation", DataType("s32")),
            Retriever("object_list_unit_id_2", DataType("s32")),
            Retriever("button_location", DataType("s32")),
            Retriever("ai_signal_value", DataType("s32")),
            Retriever("unknown_3", DataType("s32")),
            Retriever("object_attributes", DataType("s32")),
            Retriever("from_variable", DataType("s32")),
            Retriever("variable_or_timer", DataType("s32")),
            Retriever("facet", DataType("s32")),
            Retriever("location_object_reference", DataType("s32")),
            Retriever("play_sound", DataType("s32")),
            Retriever("player_color", DataType("s32")),
            Retriever("unknown_4", DataType("s32")),
            Retriever("message", DataType("str32")),
            Retriever("sound_name", DataType("str32")),
            Retriever("selected_object_ids", DataType("s32")),
        ]

        super().__init__("Effect", retrievers, parser_obj, data, pieces=pieces)

    @staticmethod
    def defaults(pieces):
        defaults = {
            'effect_type': 0,
            'static_value_46': 48,
            'ai_script_goal': -1,
            'quantity': -1,
            'aa_quantity': -1,
            'aa_armor_or_attack_type': -1,
            'tribute_list': -1,
            'diplomacy': -1,
            'number_of_units_selected': -1,
            'unknown': -1,
            'object_list_unit_id': -1,
            'source_player': -1,
            'target_player': -1,
            'technology': -1,
            'string_id': -1,
            'unknown_2': -1,
            'display_time': -1,
            'trigger_id': -1,
            'location_x': -1,
            'location_y': -1,
            'area_1_x': -1,
            'area_1_y': -1,
            'area_2_x': -1,
            'area_2_y': -1,
            'object_group': -1,
            'object_type': -1,
            'instruction_panel_position': -1,
            'attack_stance': -1,
            'time_unit': -1,
            'enabled_or_victory': -1,
            'food': -1,
            'wood': -1,
            'stone': -1,
            'gold': -1,
            'item_id': -1,
            'flash_object': -1,
            'force_research_technology': -1,
            'visibility_state': -1,
            'scroll': -1,
            'operation': -1,
            'object_list_unit_id_2': -1,
            'button_location': -1,
            'ai_signal_value': -1,
            'unknown_3': -1,
            'object_attributes': -1,
            'from_variable': -1,
            'variable_or_timer': -1,
            'facet': -1,
            'location_object_reference': -1,
            'play_sound': -1,
            'player_color': -1,
            'unknown_4': -1,
            'message': '',
            'sound_name': '',
            'selected_object_ids': [],
        }
        return defaults



from typing import Dict

from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.helper.retriever_dependency import RetrieverDependency, DependencyAction, DependencyEval, \
    DependencyTarget
from AoE2ScenarioParser.pieces.structs.aoe2_struct import AoE2Struct


class PlayerDataThreeStruct(AoE2Struct):
    dependencies: Dict[str, Dict[str, RetrieverDependency]] = {
        "player_count_for_diplomacy": {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_VALUE, DependencyTarget("self", "diplomacy_for_interaction"),
                DependencyEval("len(x)"))
        },
        "diplomacy_for_interaction": {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_REPEAT, DependencyTarget("self", "player_count_for_diplomacy")),
            "on_construct": RetrieverDependency(DependencyAction.REFRESH_SELF),
            "on_commit": RetrieverDependency(
                DependencyAction.REFRESH, DependencyTarget("self", "player_count_for_diplomacy"))
        },
        "victory_version": {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_VALUE, DependencyTarget("self", "unknown_2"),
                DependencyEval("2 if len(x) == 7 else 0"))
        },
        "unknown": {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_VALUE, DependencyTarget("self", "unknown_structure_grand_theft_empires"),
                DependencyEval("len(x)"))
        },
        "unknown_2": {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_REPEAT, DependencyTarget("self", "victory_version"),
                DependencyEval("7 if x == 2 else 0")),
            "on_construct": RetrieverDependency(DependencyAction.REFRESH_SELF),
            "on_commit": RetrieverDependency(
                DependencyAction.REFRESH, DependencyTarget("self", "victory_version"))
        },
        "unknown_5": {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_REPEAT, DependencyTarget("self", "victory_version"),
                DependencyEval("1 if x == 2 else 0")),
            "on_construct": RetrieverDependency(DependencyAction.REFRESH_SELF)
        },
        "unknown_structure_grand_theft_empires": {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_REPEAT, DependencyTarget("self", "unknown")
            ),
            "on_construct": RetrieverDependency(DependencyAction.REFRESH_SELF),
            "on_commit": RetrieverDependency(
                DependencyAction.REFRESH, DependencyTarget("self", "unknown"))
        },
        "unknown_structure_ww_campaign_2": {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_REPEAT, DependencyTarget("self", "unknown_5")),
            "on_construct": RetrieverDependency(DependencyAction.REFRESH_SELF)
        },
    }

    def __init__(self, parser_obj=None, data=None, pieces=None):
        retrievers = [
            Retriever("constant_name", DataType("str16")),
            Retriever("initial_camera_x", DataType("f32")),
            Retriever("initial_camera_y", DataType("f32")),
            Retriever("unknown_similar_to_camera_x", DataType("s16")),
            Retriever("unknown_similar_to_camera_y", DataType("s16")),
            Retriever("aok_allied_victory", DataType("u8")),
            Retriever("player_count_for_diplomacy", DataType("u16")),
            Retriever("diplomacy_for_interaction", DataType("u8")),
            Retriever("diplomacy_for_ai_system", DataType("u32", repeat=9)),
            Retriever("color", DataType("u32")),
            Retriever("victory_version", DataType("f32")),
            Retriever("unknown", DataType("u16")),
            Retriever("unknown_2", DataType("u8")),
            Retriever("unknown_structure_grand_theft_empires", DataType("44")),
            Retriever("unknown_5", DataType("u8"), possibly_list=False),
            Retriever("unknown_3", DataType("u8", repeat=7)),
            # Todo: Has to change value of unknown_5. But currently not supported
            Retriever("unknown_structure_ww_campaign_2", DataType("32")),
            Retriever("unknown_4", DataType("s32")),
        ]

        super().__init__("Player Data #3", retrievers, parser_obj, data, pieces=pieces)

    @staticmethod
    def defaults(pieces):
        defaults = {
            'constant_name': 'Scenario Editor Phantom',
            'initial_camera_x': 72.0,
            'initial_camera_y': 72.0,
            'unknown_similar_to_camera_x': 72,
            'unknown_similar_to_camera_y': 72,
            'aok_allied_victory': 0,
            'player_count_for_diplomacy': 9,
            'diplomacy_for_interaction': [3, 0, 3, 3, 3, 3, 3, 3, 3],
            'diplomacy_for_ai_system': [0, 1, 4, 4, 4, 4, 4, 4, 4],
            'color': 0,
            'victory_version': 2.0,
            'unknown': 0,
            'unknown_2': [0, 0, 0, 0, 0, 0, 0],
            'unknown_5': 0,
            'unknown_structure_grand_theft_empires': [],
            'unknown_3': [0, 0, 0, 0, 0, 0, 0],
            'unknown_structure_ww_campaign_2': [],
            'unknown_4': -1,
        }
        return defaults



from __future__ import annotations

from collections import OrderedDict
from copy import deepcopy
from typing import List, Type, TYPE_CHECKING

from AoE2ScenarioParser.helper import helper

if TYPE_CHECKING:
    from typing import OrderedDict as OrderedDictType
    from AoE2ScenarioParser.helper.retriever_object_link import RetrieverObjectLink
    from AoE2ScenarioParser.pieces.aoe2_piece import AoE2Piece


class AoE2Object:
    _link_list: List[RetrieverObjectLink] = []

    def __init__(self, **kwargs):
        self._instance_number_history = []
        self._pieces: OrderedDictType[str, AoE2Piece] = OrderedDict()

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            if k not in ['_pieces']:
                val = deepcopy(v)
            else:
                val = getattr(self, k)
            setattr(result, k, val)
        return result

    @classmethod
    def _construct(cls, pieces: OrderedDictType[str, AoE2Piece], instance_number_history=None):
        if instance_number_history is None:
            instance_number_history = []

        object_parameters: dict = {}
        for link in cls._link_list:
            object_parameters[link.name] = link.construct(pieces, instance_number_history)

        obj = cls(**object_parameters)
        obj._pieces = pieces
        return obj

    def commit(self, pieces=None, local_link_list=None):
        """
        Commits all changes to the piece & struct structure of the object it's called upon.

        Args:
            pieces (OrderedDictType[str, AoE2Piece]): A list of pieces to reference where to commit to. If left empty,
                the pieces default to the pieces where this object was constructed from. When this object wasn't present
                when the file was read it's not possible to use the
            local_link_list (Type[List[RetrieverObjectLink]]): a separate list of RetrieverObjectLinks. This way it's
                possible to commit only specific properties instead of all from an object.
        """
        if self._pieces == {} and pieces is None:
            raise ValueError("Unable to commit object. No reference to pieces set.")

        if local_link_list is None:
            local_link_list = self._link_list
        if pieces is None:
            pieces = self._pieces

        for link in local_link_list[::-1]:
            link.commit(pieces, host_obj=self)

    @staticmethod
    def get_instance_number(obj: AoE2Object = None, instance_number_history=None) -> int:
        if obj is None and instance_number_history is None:
            raise ValueError("The use of the parameter 'obj' or 'instance_number_history' is required.")
        if obj is not None and instance_number_history is not None:
            raise ValueError("Cannot use both the parameter 'obj' and 'instance_number_history'.")

        if instance_number_history is None and obj is not None:
            instance_number_history = obj._instance_number_history
        return instance_number_history[-1] if len(instance_number_history) > 0 else None

    def __repr__(self):
        self_dict = self.__dict__
        self_dict['_pieces'] = f"OrderDict"
        return str(self.__class__.__name__) + ": " + helper.pretty_print_dict(self_dict)



import math
from typing import Dict

from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.helper.retriever_dependency import RetrieverDependency, DependencyAction, DependencyTarget, \
    DependencyEval
from AoE2ScenarioParser.pieces import aoe2_piece
from AoE2ScenarioParser.pieces.structs.terrain import TerrainStruct


class MapPiece(aoe2_piece.AoE2Piece):
    dependencies: Dict[str, Dict[str, RetrieverDependency]] = {
        "villager_force_drop": {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_REPEAT,
                DependencyTarget('FileHeaderPiece', 'version'),
                DependencyEval('1 if x in [\'1.37\', \'1.40\', \'1.41\'] else 0')
            ),
            "on_construct": RetrieverDependency(DependencyAction.REFRESH_SELF)
        },
        "block_humanity_team_change": {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_REPEAT,
                DependencyTarget('FileHeaderPiece', 'version'),
                DependencyEval('1 if x in [\'1.41\'] else 0')
            ),
            "on_construct": RetrieverDependency(DependencyAction.REFRESH_SELF)
        },
        "map_width": {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_VALUE,
                DependencyTarget("self", "terrain_data"),
                DependencyEval("int(sqrt(len(x)))", {'sqrt': math.sqrt})
            )
        },
        "map_height": {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_VALUE,
                DependencyTarget("self", "terrain_data"),
                DependencyEval("int(sqrt(len(x)))", {'sqrt': math.sqrt})
            )
        },
        "terrain_data": {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_REPEAT,
                DependencyTarget("self", "map_width"),
                DependencyEval("pow(x, 2)")
            ),
            "on_construct": RetrieverDependency(DependencyAction.REFRESH_SELF),
            "on_commit": RetrieverDependency(
                DependencyAction.REFRESH, DependencyTarget(["self", "self"], ["map_width", "map_height"])
            )
        },
        "script_name": {
            "on_commit": RetrieverDependency(
                DependencyAction.REFRESH, DependencyTarget("FilesPiece", "script_file_path")
            )
        },
    }

    def __init__(self, parser_obj=None, data=None, pieces=None):
        retrievers = [
            Retriever('separator_1', DataType("2")),
            Retriever('unknown_string', DataType("str16")),
            Retriever('separator_2', DataType("2")),
            Retriever('map_color_mood', DataType("str16")),
            Retriever('separator_3', DataType("2")),
            # [VERSION CHANGE] ADDED in 1.37 > 1.40
            Retriever('script_name', DataType("str16")),
            # [VERSION CHANGE] ADDED in 1.40 > 1.41
            Retriever('block_humanity_team_change', DataType("u8")),
            Retriever('collide_and_correct', DataType("u8")),
            # [VERSION CHANGE] ADDED in 1.36 > 1.37
            Retriever('villager_force_drop', DataType("u8")),
            # [VERSION CHANGE] ADDED in 1.37 > 1.40
            Retriever('unknown', DataType("128")),
            Retriever('player_1_camera_y', DataType("s32")),
            Retriever('player_1_camera_x', DataType("s32")),
            Retriever('ai_type', DataType("s8")),
            Retriever('map_width', DataType("s32")),
            Retriever('map_height', DataType("s32")),
            Retriever('terrain_data', DataType(TerrainStruct))
        ]

        super().__init__("Map", retrievers, parser_obj, data=data, pieces=pieces)

    @staticmethod
    def defaults(pieces):
        defaults = {
            'separator_1': b'`\n',
            'unknown_string': '',
            'separator_2': b'`\n',
            'map_color_mood': 'Empty',
            'separator_3': b'`\n',
            'script_name': '',
            'block_humanity_team_change': 0,
            'collide_and_correct': 0,
            'villager_force_drop': 0,
            'unknown': b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff'
                       b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff'
                       b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff'
                       b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff'
                       b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff'
                       b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff',
            'player_1_camera_y': -559026163,
            'player_1_camera_x': 2,
            'ai_type': 1,
            'map_width': 120,
            'map_height': 120,
            'terrain_data': [TerrainStruct(data=list(TerrainStruct.defaults(pieces).values()), pieces=pieces) for _ in
                             range(120 * 120)],
        }
        return defaults



from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.pieces.structs.aoe2_struct import AoE2Struct


class TerrainStruct(AoE2Struct):
    def __init__(self, parser_obj=None, data=None, pieces=None):
        retrievers = [
            Retriever("terrain_id", DataType("u8")),
            Retriever("elevation", DataType("u8")),
            Retriever("unused", DataType("u8")),
            Retriever("separator", DataType("2")),
            Retriever("layer", DataType("s16"))
        ]

        super().__init__("Terrain", retrievers, parser_obj, data, pieces=pieces)

    @staticmethod
    def defaults(pieces):
        defaults = {
            'terrain_id': 0,
            'elevation': 0,
            'unused': 0,
            'separator': b'\xff\xff',
            'layer': -1,
        }
        return defaults



import time
from typing import Dict

from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.helper.retriever_dependency import RetrieverDependency, DependencyAction, DependencyTarget, \
    DependencyEval
from AoE2ScenarioParser.pieces import aoe2_piece


class FileHeaderPiece(aoe2_piece.AoE2Piece):
    dependencies: Dict[str, Dict[str, RetrieverDependency]] = {
        # "individual_victories_used": {
        #     "on_refresh": RetrieverDependency(
        #         DependencyAction.SET_REPEAT,
        #         DependencyTarget('self', 'version'),
        #         DependencyEval('0 if x == \'1.40\' else 1')
        #     ),
        #     "on_construct": RetrieverDependency(DependencyAction.REFRESH_SELF)
        # }
        # amount_of_unknown_numbers
        # unknown_numbers
        "amount_of_unknown_numbers": {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_VALUE, DependencyTarget("self", "unknown_numbers"),
                DependencyEval("len(x)")
            )
        },
        "unknown_numbers": {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_REPEAT, DependencyTarget("self", "amount_of_unknown_numbers")
            ),
            "on_construct": RetrieverDependency(DependencyAction.REFRESH_SELF),
            "on_commit": RetrieverDependency(
                DependencyAction.REFRESH, DependencyTarget("self", "amount_of_unknown_numbers")
            )
        },
        "trigger_count": {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_VALUE, DependencyTarget("TriggerPiece", "trigger_data"),
                DependencyEval("len(x)")
            )
        },
    }

    def __init__(self, parser_obj=None, data=None, pieces=None):
        retrievers = [
            Retriever("version", DataType("c4")),
            Retriever("header_length", DataType("u32")),
            Retriever("savable", DataType("s32")),
            Retriever("timestamp_of_last_save", DataType("u32")),
            Retriever("scenario_instructions", DataType("str32")),
            # Retriever("individual_victories_used", DataType("u32")),
            Retriever("player_count", DataType("u32")),
            Retriever("unknown_value", DataType("u32")),  # Always (?) 1k
            Retriever("unknown_value_2", DataType("u32")),  # Always (?) 1
            Retriever("amount_of_unknown_numbers", DataType("u32")),
            Retriever("unknown_numbers", DataType("u32")),
            Retriever("creator_name", DataType("str32")),
            Retriever("trigger_count", DataType("u32")),
        ]

        super().__init__("File Header", retrievers, parser_obj, data=data, pieces=pieces)

    @staticmethod
    def defaults(pieces):
        defaults = {
            'version': '1.41',
            'header_length': 0,
            'savable': 6,
            'timestamp_of_last_save': int(time.time()),
            'scenario_instructions': '',
            # 'individual_victories_used': 0,
            'player_count': 2,
            'unknown_value': 1000,
            'unknown_value_2': 1,
            'amount_of_unknown_numbers': 6,
            'unknown_numbers': [2, 3, 4, 5, 6, 7],
            'creator_name': 'KSneijders/AoE2ScenarioParser',
            'trigger_count': 0,
        }
        return defaults



from AoE2ScenarioParser.helper import helper


def create_generator(uncompressed_file, chunk_size=4):
    for x in range(0, len(uncompressed_file), chunk_size):
        yield uncompressed_file[x:x + chunk_size]


def repeat_generator(generator, run_times, intended_stop_iteration=False, return_bytes=True):
    elements = b'' if return_bytes else []

    try:
        for i in range(0, run_times):
            if return_bytes:
                elements += next(generator)
            else:
                # Generator returns list. So lists are merged
                elements += next(generator)
    except StopIteration as e:
        if not intended_stop_iteration:
            print(f"\n\n[StopIteration] in repeat_generator while retrieving {run_times} bytes.")
            print(f"{' ' * 14}> Bytes are being written to ErrorFile_RepeaterBytes.txt...")
            error_file = open("../ErrorFile_RepeaterBytes.txt", 'w')
            error_file.write(helper.create_textual_hex(elements.hex(), space_distance=2, enter_distance=24))
            error_file.close()
            print(f"{' ' * 14}> Writing bytes finished.")
        raise e
    return elements



from typing import Dict

from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.helper.retriever_dependency import RetrieverDependency, DependencyAction, DependencyTarget, \
    DependencyEval
from AoE2ScenarioParser.pieces import aoe2_piece
from AoE2ScenarioParser.pieces.structs.player_data_four import PlayerDataFourStruct
from AoE2ScenarioParser.pieces.structs.player_data_three import PlayerDataThreeStruct
from AoE2ScenarioParser.pieces.structs.player_units import PlayerUnitsStruct


class UnitsPiece(aoe2_piece.AoE2Piece):
    dependencies: Dict[str, Dict[str, RetrieverDependency]] = {
        "number_of_unit_sections": {

            "on_refresh": RetrieverDependency(
                DependencyAction.SET_VALUE, DependencyTarget("self", "players_units"),
                DependencyEval('len(x)')
            )
        },
        "players_units": {
            "on_refresh": RetrieverDependency(
                DependencyAction.SET_REPEAT, DependencyTarget("self", "number_of_unit_sections")
            ),
            "on_construct": RetrieverDependency(DependencyAction.REFRESH_SELF),
            "on_commit": RetrieverDependency(
                DependencyAction.REFRESH,
                DependencyTarget("self", "number_of_unit_sections")
            )
        }
    }

    def __init__(self, parser_obj=None, data=None, pieces=None):
        retrievers = [
            Retriever("number_of_unit_sections", DataType("u32")),
            Retriever("player_data_4", DataType(PlayerDataFourStruct, repeat=8)),
            Retriever("number_of_players", DataType("u32")),  # Also always 9 (Gaia + 8Plyrs)
            Retriever("player_data_3", DataType(PlayerDataThreeStruct, repeat=8)),
            Retriever("players_units", DataType(PlayerUnitsStruct)),
        ]

        super().__init__("Units", retrievers, parser_obj, data=data, pieces=pieces)

    @staticmethod
    def defaults(pieces):
        defaults = {
            'number_of_unit_sections': 9,
            'player_data_4': [PlayerDataFourStruct(data=[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 200.0], pieces=pieces) for _ in
                              range(8)],
            'number_of_players': 9,
            'player_data_3': UnitsPiece._get_player_data_three_default(pieces),
            'players_units': [PlayerUnitsStruct(data=[0, []], pieces=pieces) for _ in range(9)],
        }
        return defaults

    @staticmethod
    def _get_player_data_three_default(pieces):
        data = [list(PlayerDataThreeStruct.defaults(pieces).values()) for _ in range(8)]
        for i in range(len(data)):
            # diplomacy_for_interaction
            data[i][7] = [3] + ([3] * i) + [0] + ([3] * (7 - i))
            # diplomacy_for_ai_system
            data[i][8] = [0] + ([4] * i) + [1] + ([4] * (7 - i))
            # Color
            data[i][9] = i
        return [PlayerDataThreeStruct(data=d, pieces=pieces) for d in data]



from enum import Enum


class RetrieverDependency:
    def __init__(self, dependency_type, dependency_target=None, dependency_eval=None):
        """
        Object for handling dependencies between retrievers

        Args:
            dependency_type (DependencyAction): TODO: Add docstring
            dependency_target (DependencyTarget): TODO: Add docstring
            dependency_eval (DependencyEval): TODO: Add docstring
        """

        self.dependency_type = dependency_type
        self.dependency_target = dependency_target

        if dependency_type in [DependencyAction.SET_REPEAT, DependencyAction.SET_VALUE] and dependency_eval is None:
            dependency_eval = DependencyEval('x')

        self.dependency_eval = dependency_eval

    def __repr__(self):
        return f"[RetrieverDependency] {self.dependency_type} " \
               f"\n\tTarget: {self.dependency_target}" \
               f"\n\tEval: {self.dependency_eval.eval_code}"


class DependencyTarget:
    def __init__(self, target_piece, piece_attr_name):
        """
        Object for targeting a specific retriever based on it's piece name (or 'self') and the retriever name.

        Args:
            target_piece (Union[str, List[str]]): The name of the targeted piece or 'self' for own list
            piece_attr_name (Union[str, List[str]]): The name of the attribute in targeted piece or own list
        """
        if not ((type(target_piece) is list and type(piece_attr_name) is list) or
                (type(target_piece) is str and type(piece_attr_name) is str)):
            raise TypeError("Both parameters should be of the same type. This can be list or str.")
        if type(target_piece) is list:
            if not (len(target_piece) == len(piece_attr_name)):
                raise ValueError("Both parameters should be of the same length when using lists.")
        self.target_piece = target_piece
        self.piece_attr_name = piece_attr_name

    def __repr__(self) -> str:
        return f"[DependencyTarget] {self.target_piece} -> {self.piece_attr_name}"


class DependencyEval:
    def __init__(self, eval_code, eval_locals=None, values_as_variable=None):
        """
        Object for storing dependency eval code and it's locals.

        Args:
            eval_code (str): The code executed using eval
            eval_locals (dict): The locals dict handed to the eval function
        """
        if eval_locals is None:
            eval_locals = {}
        if values_as_variable is None:
            values_as_variable = []
        self.eval_code = eval_code
        self.eval_locals = eval_locals
        self.values_as_variable = values_as_variable


class DependencyAction(Enum):
    REFRESH = 0
    REFRESH_SELF = 1
    SET_REPEAT = 2
    SET_VALUE = 3



from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.pieces.structs.aoe2_struct import AoE2Struct


class UnitStruct(AoE2Struct):
    def __init__(self, parser_obj=None, data=None, pieces=None):
        retrievers = [
            Retriever("x", DataType("f32")),
            Retriever("y", DataType("f32")),
            Retriever("z", DataType("f32")),
            Retriever("reference_id", DataType("s32")),
            Retriever("unit_const", DataType("u16")),
            Retriever("status", DataType("u8")),
            # Status, Always 2. 0-6 no difference (?) | 7-255 makes it disappear. (Except from the mini-map)
            Retriever("rotation", DataType("f32")),
            Retriever("initial_animation_frame", DataType("u16")),
            Retriever("garrisoned_in_id", DataType("s32")),
        ]

        super().__init__("Unit", retrievers, parser_obj, data, pieces=pieces)

    @staticmethod
    def defaults(pieces):
        defaults = {
            'x': 0.5,
            'y': 0.5,
            'z': 0,
            'reference_id': 0,
            'unit_const': 4,  # Archer
            'status': 2,
            'rotation': 0,
            'initial_animation_frame': 0,
            'garrisoned_in_id': -1,
        }
        return defaults



from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.pieces.structs.aoe2_struct import AoE2Struct


class PlayerDiplomacyStruct(AoE2Struct):
    def __init__(self, parser_obj=None, data=None, pieces=None):
        retrievers = [
            Retriever("stance_with_each_player", DataType("u32", repeat=16)),
        ]

        super().__init__("Player Diplomacy", retrievers, parser_obj, data, pieces=pieces)

    @staticmethod
    def defaults(pieces):
        defaults = {
            'stance_with_each_player': [3] * 16
        }
        return defaults



from typing import Type, List

from AoE2ScenarioParser.helper.parser import handle_retriever_dependency
from AoE2ScenarioParser.helper.retriever import get_retriever_by_name
from AoE2ScenarioParser.objects.aoe2_object import AoE2Object


class RetrieverObjectLink:
    def __init__(self,
                 variable_name: str,
                 piece: str = None,
                 link: str = None,
                 process_as_object: Type[AoE2Object] = None,
                 retrieve_instance_number: bool = False,
                 retrieve_history_number=-1
                 ):
        if (link is not None) + retrieve_instance_number + (retrieve_history_number != -1) != 1:
            raise ValueError("Use one and only one of the following parameters: 'link', 'retrieve_instance_number' or "
                             "'retrieve_history_number'.")
        self.name: str = variable_name
        self.piece = piece
        self.link = link
        self.is_special_unit_case = self._self_is_special_unit_case()
        self.process_as_object: Type[AoE2Object] = process_as_object
        self.retrieve_instance_number: bool = retrieve_instance_number
        self.retrieve_history_number: int = retrieve_history_number

    def construct(self, pieces, instance_number_history=None):
        if instance_number_history is None:
            instance_number_history = []

        instance_number = AoE2Object.get_instance_number(instance_number_history=instance_number_history)

        if self.retrieve_instance_number:
            return instance_number
        elif self.retrieve_history_number != -1:
            return instance_number_history[self.retrieve_history_number]
        else:
            if self.is_special_unit_case:
                return self._construct_special_unit_case(pieces)
            # Use temp_link to not change the actual links as they are class attributes
            temp_link = self.link
            if instance_number_history:
                for i in instance_number_history:
                    temp_link = temp_link.replace("__index__", str(i), 1)
            temp_link = temp_link.replace("__index__", str(instance_number), 1)

            value = pieces[self.piece]
            for x in temp_link.split("."):
                if "[" in x:
                    indexing = x.index('[')
                    value = getattr(value, x[:indexing])
                    value = value[int(x[indexing + 1:len(x) - 1])]
                else:
                    value = getattr(value, x)

            if self.process_as_object is not None:
                value_list = []
                for index, struct in enumerate(value):
                    value_list.append(
                        self.process_as_object._construct(
                            pieces,
                            instance_number_history=instance_number_history + [index]
                        )
                    )
                return value_list
            return value

    def commit(self, pieces, host_obj):
        # Object-only retrievers for the ease of access of information.
        # Not actually representing a value in the scenario file.
        if self.retrieve_instance_number or self.retrieve_history_number >= 0:
            return

        instance_number_history = host_obj._instance_number_history
        instance_number = AoE2Object.get_instance_number(obj=host_obj)
        value = host_obj.__getattribute__(self.name)

        # Replace __index__ values
        temp_link = self.link
        if instance_number_history:
            for i in instance_number_history:
                temp_link = temp_link.replace("__index__", str(i), 1)
        temp_link = temp_link.replace("__index__", str(instance_number), 1)

        if self.is_special_unit_case:
            temp_link = temp_link.replace('[]', '[0]', 1)

        split_temp_link: List[str] = temp_link.split(".")
        retriever = None

        piece = pieces[self.piece]
        for attribute in split_temp_link:
            if '[' in attribute:
                index_location = attribute.index('[')
                index = int(attribute[index_location + 1:len(attribute) - 1])
                piece = get_retriever_by_name(piece.retrievers, attribute[:index_location]).data[index]
            else:
                retriever = get_retriever_by_name(piece.retrievers, attribute)

        if retriever is None:
            raise ValueError("RetrieverObjectLink is unable to connect to retriever")

        retriever_list = piece.retrievers

        if self.process_as_object is not None:
            link_piece = retriever.datatype.var

            if self.is_special_unit_case:
                self._commit_special_unit_case(host_obj, pieces, link_piece, value)
                return

            try:
                old_length = len(retriever.data)
            except TypeError:
                # retriever.data was not set before (list of 0 -> None)
                retriever.data = []
                old_length = 0
            new_length = len(value)

            if new_length < old_length:
                retriever.data = retriever.data[:new_length]
            elif new_length > old_length:
                retriever.data += [link_piece() for x in range(new_length - old_length)]
                if retriever.log_value:
                    retriever._update_print(
                        f"[{link_piece.__name__}] * {old_length}",
                        f"[{link_piece.__name__}] * {new_length}"
                    )

            for index, obj in enumerate(value):
                obj._pieces = pieces
                obj._instance_number_history = host_obj._instance_number_history + [index]
                obj.commit()
        else:
            retriever.data = value

        if hasattr(retriever, 'on_commit'):
            handle_retriever_dependency(retriever, retriever_list, "commit", pieces)

    def _self_is_special_unit_case(self):
        if self.link is not None:
            return "[]" in self.link
        return False

    def _construct_special_unit_case(self, pieces):
        list_notation_location = self.link.find("[]")
        struct_list = getattr(pieces[self.piece], self.link[:list_notation_location])
        list_len = len(struct_list)
        result_list = [[] for _ in range(list_len)]
        for i in range(list_len):
            list_of_attributes = struct_list[i].__getattr__(self.link[list_notation_location + 3:])
            for j in range(len(list_of_attributes)):
                result_list[i].append(self.process_as_object._construct(pieces, [i, j]))
        return result_list

    def _commit_special_unit_case(self, host_obj, pieces, link_piece, units):
        for player_number in range(len(units)):
            pieces['UnitsPiece'].players_units[player_number].unit_count = len(units[player_number])
            pieces['UnitsPiece'].players_units[player_number].units = \
                [link_piece() for _ in range(len(units[player_number]))]

            for index, obj in enumerate(units[player_number]):
                obj._pieces = pieces
                obj._instance_number_history = [player_number, index]
                obj.commit()

    def __repr__(self):
        return "[RetrieverObjectLink] " + self.name + ": " + str(self.piece) + "." + str(self.link) + \
               (f"\n\t- Process as: {self.process_as_object.__name__}" if self.process_as_object else "") + \
               (f"\n\t- Get Instance Number: True" if self.retrieve_instance_number else "") + \
               (f"\n\t- Get Hist Number: {self.retrieve_history_number}" if self.retrieve_history_number >= 0 else "")



from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.pieces.structs.aoe2_struct import AoE2Struct


class AI2Struct(AoE2Struct):
    def __init__(self, parser_obj=None, data=None, pieces=None):
        retrievers = [
            Retriever("ai_file_name", DataType("str32")),
            Retriever("ai_file", DataType("str32")),
        ]

        super().__init__("AI2", retrievers, parser_obj, data, pieces=pieces)

    @staticmethod
    def defaults(pieces):
        defaults = {
            'ai_file_name': '',
            'ai_file': '',
        }
        return defaults



from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.pieces import aoe2_piece
from AoE2ScenarioParser.pieces.structs.ai import AIStruct
from AoE2ScenarioParser.pieces.structs.resources import ResourcesStruct


class PlayerDataTwoPiece(aoe2_piece.AoE2Piece):
    def __init__(self, parser_obj=None, data=None, pieces=None):
        retrievers = [
            Retriever("strings", DataType("str16", repeat=32)),
            Retriever("ai_names", DataType("str16", repeat=16)),
            Retriever("ai_files", DataType(AIStruct, repeat=16)),
            Retriever("ai_type", DataType("u8", repeat=16)),
            Retriever("separator", DataType("u32")),
            Retriever("resources", DataType(ResourcesStruct, repeat=16))
        ]

        super().__init__("Player Data #2", retrievers, parser_obj, data=data, pieces=pieces)

    @staticmethod
    def defaults(pieces):
        defaults = {
            'strings': [''] * 32,
            'ai_names': [''] * 16,
            'ai_files': [AIStruct(data=list(AIStruct.defaults(pieces).values()), pieces=pieces) for _ in range(16)],
            'ai_type': [1] * 16,
            'separator': 4294967197,
            'resources': PlayerDataTwoPiece._get_resource_struct_default(pieces),
        }
        return defaults

    @staticmethod
    def _get_resource_struct_default(pieces):
        data = [list(ResourcesStruct.defaults(pieces).values()) for _ in range(16)]
        for i in range(len(data)):
            data[i][5] = i
        return [ResourcesStruct(data=x, pieces=pieces) for x in data]
